@@ -1,3 +1,9 @@
+/* 
+ * (C) Copyright 1992, ..., 1999 the "DOSEMU-Development-Team".
+ *
+ * for details see file COPYING in the DOSEMU distribution
+ */
+
 #include "config.h"
 
 #ifdef USE_MHPDBG
@@ -7,10 +13,6 @@
 /* Define if we want graphics in X (of course we want :-) (root@zaphod) */
 /* WARNING: This may not work in BSD, because it was written for Linux! */
 
-#ifndef lint
-static char rcsid[]="$Id: sigsegv.c,v 2.20 1995/04/08 22:30:40 root Exp $";
-#endif /* lint */
-
 #include <stdio.h>
 #include <termios.h>
 #include <stdlib.h>
@@ -20,7 +22,6 @@ static char rcsid[]="$Id: sigsegv.c,v 2.20 1995/04/08 22:30:40 root Exp $";
 #include <sys/times.h>
 #include <sys/time.h>
 #include <sys/types.h>
-#include <sys/times.h>
 #if X_GRAPHICS
 #include <sys/mman.h>           /* root@sjoerd*/
 #endif /* X_GRAPHICS */
@@ -123,10 +124,6 @@ dosemu_fault(int signal, int code, struct sigcontext *scp)
 }
 #endif /* __NetBSD__ */
 
-#ifdef __linux__
-#define dosemu_fault1 dosemu_fault
-#endif /* __linux__ */
-
 
 
 /* Function prototypes */
@@ -146,16 +143,13 @@ void print_exception_info(struct sigcontext_struct *scp);
 void 
 dosemu_fault1(
 #ifdef __linux__
-int signal, struct sigcontext_struct context
+int signal, struct sigcontext_struct *scp
 #endif /* __linux__ */
 #ifdef __NetBSD__
 int signal, int code, struct sigcontext *scp
 #endif /* __NetBSD__ */
 )
 {
-#ifdef __linux__
-  struct sigcontext_struct *scp = &context;
-#endif /* __linux__ */
   unsigned char *csp;
   int i;
 
@@ -169,12 +163,39 @@ int signal, int code, struct sigcontext *scp
       case 0x05: /* bounds */
       case 0x07: /* device_not_available */
 		 return (void) do_int(_trapno);
+
+      case 0x10: /* coprocessor error */
+		 pic_request(PIC_IRQ13); /* this is the 386 way of signalling this */
+		 return;
+
+      case 0x11: /* alignment check */
+		 /*
+		  * FIRST thing to do - to avoid being trapped into int0x11
+		  * forever, we must clear AC before doing anything else!
+		  */
+		 __asm__ __volatile__ ("
+			pushfl
+			popl	%%eax
+			andl	$0xfffbffff,%%eax
+			pushl	%%eax
+			popfl"
+			: : : "%eax");
+		 /* we are now safe; nevertheless, fall into the default
+		  * case and exit dosemu, as an AC fault in vm86 is(?) a
+		  * catastrophic failure.
+		  */
+		 goto sgleave;
+
       case 0x06: /* invalid_op */
 		 dbug_printf("SIGILL while in vm86()\n");
 #if 0
 		 show_regs(__FILE__, __LINE__);
 #endif /* 0 */
  		 csp = SEG_ADR((unsigned char *), cs, ip);
+#if 1
+		 /* this one is for CPU detection programs */
+		 if (csp[0]==0x0f) return (void) do_int(_trapno);
+#endif
  		 /* Some db commands start with 2e (use cs segment) 
 		    and thus is accounted for here */
  		 if (csp[0] == 0x2e) {
@@ -203,9 +224,9 @@ int signal, int code, struct sigcontext *scp
 #endif /* X_GRAPHICS */
 
       default:	
-
+sgleave:
 #if 0
-		 error("ERROR: unexpected CPU exception 0x%02lx errorcode: 0x%08lx while in vm86()\n"
+		 error("unexpected CPU exception 0x%02lx errorcode: 0x%08lx while in vm86()\n"
 	  	"eip: 0x%08lx  esp: 0x%08lx  eflags: 0x%lx\n"
 	  	"cs: 0x%04x  ds: 0x%04x  es: 0x%04x  ss: 0x%04x\n", _trapno,
 		_err,
@@ -214,7 +235,7 @@ int signal, int code, struct sigcontext *scp
 
 		 print_exception_info(scp);
 #else
-		 error("ERROR: unexpected CPU exception 0x%02lx errorcode: 0x%08lx while in vm86 (DOS)\n",
+		 error("unexpected CPU exception 0x%02lx errorcode: 0x%08lx while in vm86 (DOS)\n",
 	  	 _trapno, _err);
 		{
 		  extern FILE *dbg_fd;
@@ -239,10 +260,10 @@ int signal, int code, struct sigcontext *scp
 #if X_GRAPHICS				/* only for debug ?*/ /*root@sjoerd*/
 /* The reason it comes here instead of inside_VM86 is
  * char_out() in ./video/int10.c (and the memory is protected).
+ * We want to protect the video memory and the VGA BIOS.
  * This function is called from the following functions:
  * dosemu/utilities.c:     char_out(*s++, READ_BYTE(BIOS_CURRENT_SCREEN_PAGE));
  * video/int10.c:    char_out(*(char *) &REG(eax), READ_BYTE(BIOS_CURRENT_SCREEN_PAGE));
- *
  */
 
   if(_trapno==0x0e)
@@ -280,7 +301,7 @@ int signal, int code, struct sigcontext *scp
 #else
   {
 #endif /* 0 */
-    error("ERROR: cpu exception in dosemu code outside of VM86()!\n"
+    error("cpu exception in dosemu code outside of VM86()!\n"
 	  "trapno: 0x%02lx  errorcode: 0x%08lx  cr2: 0x%08lx\n"
 	  "eip: 0x%08lx  esp: 0x%08lx  eflags: 0x%08lx\n"
 	  "cs: 0x%04x  ds: 0x%04x  es: 0x%04x  ss: 0x%04x\n",
@@ -341,6 +362,12 @@ int signal, int code, struct sigcontext *scp
   }
 }
 
+#ifdef __linux__
+void dosemu_fault(int signal, struct sigcontext_struct context)
+{
+    dosemu_fault1 (signal, &context);
+}
+#endif /* __linux__ */
 
 /*
  * DANG_BEGIN_FUNCTION print_exception_info
@@ -359,45 +386,45 @@ void print_exception_info(struct sigcontext_struct *scp)
   switch(_trapno)
     {
     case 0:
-      error("Division by zero\n");
+      error("@Division by zero\n");
       break;
 
 
     case 1:
-      error("Debug exception\n");
+      error("@Debug exception\n");
       break;
 
 
     case 3:
-      error("Breakpoint exception (caused by INT 3 instruction)\n");
+      error("@Breakpoint exception (caused by INT 3 instruction)\n");
       break;
 
 
     case 4:
-      error("Overflow exception (caused by INTO instruction)\n");
+      error("@Overflow exception (caused by INTO instruction)\n");
       break;
 
 
     case 5:
-      error("Bound exception (caused by BOUND instruction)\n");
+      error("@Bound exception (caused by BOUND instruction)\n");
       break;
 
 
     case 6:
-      error("Invalid opcode\n");
-      error("Opcodes: ");
+      error("@Invalid opcode\n");
+      error("@Opcodes: ");
       csp = (unsigned char *) _eip - 10;
       for (i = 0; i < 10; i++)
-	error("%02x ", *csp++);
-      error("-> ");
+	error("@%02x ", *csp++);
+      error("@-> ");
       for (i = 0; i < 10; i++)
-	error("%02x ", *csp++);
-      error("\n");    
+	error("@%02x ", *csp++);
+      error("@\n");    
       break;
 
 
     case 7:
-      error("Coprocessor exception (coprocessor not available)\n");
+      error("@Coprocessor exception (coprocessor not available)\n");
       /* I'd like to print some info on the EM, MP, and TS flags in CR0,
        * but I don't know where I can get that information :-(
        * Anyway, this exception should not happen... (Erik Mouw)
@@ -406,102 +433,102 @@ void print_exception_info(struct sigcontext_struct *scp)
 
 
     case 8:
-      error("Double fault\n");
+      error("@Double fault\n");
       break;
 
 
     case 9:
-      error("Coprocessor segment overflow\n");
+      error("@Coprocessor segment overflow\n");
       break;
 
 
     case 0xa:
-      error("Invalid TSS\n");
-      if(_err && 0x02)
-	error("IDT");
-      else if(_err && 0x04)
-	error("LDT");
+      error("@Invalid TSS\n");
+      if(_err & 0x02)
+	error("@IDT");
+      else if(_err & 0x04)
+	error("@LDT");
       else
-	error("GDT");
+	error("@GDT");
 
-      error(" selector: 0x%04x\n", ((_err >> 3) && 0x1fff ));
+      error("@ selector: 0x%04x\n", ((_err >> 3) & 0x1fff ));
 
-      if(_err && 0x01)
-	error("Exception was not caused by DOSEMU\n");
+      if(_err & 0x01)
+	error("@Exception was not caused by DOSEMU\n");
       else
-	error("Exception was caused by DOSEMU\n");
+	error("@Exception was caused by DOSEMU\n");
       break;
 
 
     case 0xb:
-      error("Segment not available\n");
+      error("@Segment not available\n");
       /* This is the same code as case 0x0a; the compiler merges these
        * blocks, so I don't have to edit some dirty constructions to
        * generate one block of code. (Erik Mouw)
        */
-      if(_err && 0x02)
-	error("IDT");
-      else if(_err && 0x04)
-	error("LDT");
+      if(_err & 0x02)
+	error("@IDT");
+      else if(_err & 0x04)
+	error("@LDT");
       else
-	error("GDT");
+	error("@GDT");
 
-      error(" selector: 0x%04x\n", ((_err >> 3) && 0x1fff ));
+      error("@ selector: 0x%04x\n", ((_err >> 3) & 0x1fff ));
 
-      if(_err && 0x01)
-	error("Exception was not caused by DOSEMU\n");
+      if(_err & 0x01)
+	error("@Exception was not caused by DOSEMU\n");
       else
-	error("Exception was caused by DOSEMU\n");
+	error("@Exception was caused by DOSEMU\n");
       break;
 
 
     case 0xc:
-      error("Stack exception\n");
+      error("@Stack exception\n");
       break;
 
 
     case 0xd:
-      error("General protection exception\n");
+      error("@General protection exception\n");
       /* This is the same code as case 0x0a; the compiler merges these
        * blocks, so I don't have to edit some dirty constructions to
        * generate one block of code. (Erik Mouw)
        */
-      if(_err && 0x02)
-	error("IDT");
-      else if(_err && 0x04)
-	error("LDT");
+      if(_err & 0x02)
+	error("@IDT");
+      else if(_err & 0x04)
+	error("@LDT");
       else
-	error("GDT");
+	error("@GDT");
 
-      error(" selector: 0x%04x\n", ((_err >> 3) && 0x1fff ));
+      error("@ selector: 0x%04x\n", ((_err >> 3) & 0x1fff ));
 
-      if(_err && 0x01)
-	error("Exception was not caused by DOSEMU\n");
+      if(_err & 0x01)
+	error("@Exception was not caused by DOSEMU\n");
       else
-	error("Exception was caused by DOSEMU\n");
+	error("@Exception was caused by DOSEMU\n");
       break;
 
 
     case 0xe:
-      error("Page fault: ");
-      if(_err && 0x02)
-	error("write");
+      error("@Page fault: ");
+      if(_err & 0x02)
+	error("@write");
       else
-	error("read");
+	error("@read");
 
-      error(" instruction to linear address: 0x%08lx\n", _cr2);
+      error("@ instruction to linear address: 0x%08lx\n", _cr2);
 
-      error("CPU was in ");
-      if(_err && 0x04)
-	error("user mode\n");
+      error("@CPU was in ");
+      if(_err & 0x04)
+	error("@user mode\n");
       else
-	error("supervisor mode\n");
+	error("@supervisor mode\n");
 
-      error("Exception was caused by ");
-      if(_err && 0x01)
-	error("insufficient privelege\n");
+      error("@Exception was caused by ");
+      if(_err & 0x01)
+	error("@insufficient privelege\n");
       else
-	error("non-available page\n");
+	error("@non-available page\n");
       break;
 
    case 0x10:
@@ -509,7 +536,7 @@ void print_exception_info(struct sigcontext_struct *scp)
       break;
 
     default:
-      error("Unknown exception\n");
+      error("@Unknown exception\n");
       break;
     }
 }
