@@ -1,3 +1,10 @@
+/* 
+ * All modifications in this file to the original code are
+ * (C) Copyright 1992, ..., 1999 the "DOSEMU-Development-Team".
+ *
+ * for details see file COPYING in the DOSEMU distribution
+ */
+
 /*
  * DOSEMU debugger,  1995 Max Parke <mhp@light.lightlink.com>
  *
@@ -41,20 +48,25 @@
 #include <linux/unistd.h>
 #include <ctype.h>
 #include <sys/types.h>
+#include <sys/stat.h>
+#include <fcntl.h>
+#include <errno.h>
 #include <sys/ioctl.h>
 #include <sys/socket.h>
 #include <netinet/in.h>
-#include <asm/bitops.h>
+#include <stdarg.h>
+#include <regex.h>
+
+#include "bitops.h"
 #include "config.h"
 #include "emu.h"
 #include "cpu.h"
 #include "dpmi.h"
+#include "utilities.h"
 
 #define MHP_PRIVATE
 #include "mhpdbg.h"
 
-#define MAP_LOC		"/usr/src/dosemu/bin/dosemu.map"
-
 #define makeaddr(x,y) ((((unsigned long)x) << 4) + (unsigned long)y)
 
 #if 0 /* modify_ldt already defined in dpmi.c */
@@ -65,20 +77,21 @@ extern int modify_ldt(int func, void *ptr, unsigned long bytecount);
 /* externs */
 extern struct mhpdbgc mhpdbgc;
 extern int a20;
+extern void mhp_send(void), mhp_close(void), mhp_putc(char c1);
 
 #if 0
 /* NOTE: the below is already defined with #include "emu.h"
- *       Must NOT redefine it, else USE_VM86PLUS won't work !!!
+ *       Must NOT redefine it, else REQUIRES_VM86PLUS won't work !!!
  */
 extern struct vm86_struct vm86s;
 #endif
 
 extern int  dis_8086(unsigned int, const unsigned char *,
-                     unsigned char *, int, unsigned int *, unsigned int *, unsigned int);
+                     unsigned char *, int, unsigned int *, unsigned int *,
+                     unsigned int, int);
 
 /* prototypes */
 static void* mhp_getadr(unsigned char *, unsigned int *, unsigned int *, unsigned int *);
-static int mhp_parse(char*, char *[]);
 static void mhp_regs  (int, char *[]);
 static void mhp_r0    (int, char *[]);
 static void mhp_dis   (int, char *[]);
@@ -104,6 +117,12 @@ static void mhp_kill    (int, char *[]);
 static void mhp_help    (int, char *[]);
 static void mhp_enter   (int, char *[]);
 static void mhp_print_ldt       (int, char *[]);
+static void mhp_debuglog (int, char *[]);
+static void mhp_dump_to_file (int, char *[]);
+static void mhp_bplog   (int, char *[]);
+static void mhp_bclog   (int, char *[]);
+static void print_log_breakpoints(void);
+
 static unsigned int lookup(unsigned char *, unsigned int *, unsigned int *);
 static inline int get_ldt(void *);
 
@@ -137,6 +156,7 @@ static const struct cmd_db cmdtab[] = {
    {"r0",            mhp_r0},
    {"r" ,            mhp_regs},
    {"e",             mhp_enter},
+   {"ed",            mhp_enter},
    {"d",             mhp_dis},
    {"u",             mhp_disasm},
    {"g",             mhp_go},
@@ -154,11 +174,15 @@ static const struct cmd_db cmdtab[] = {
    {"bpintd",        mhp_bpintd},
    {"bcintd",        mhp_bcintd},
    {"bpload",        mhp_bpload},
+   {"bplog",         mhp_bplog},
+   {"bclog",         mhp_bclog},
    {"rmapfile",      mhp_rmapfile},
    {"rusermap",      mhp_rusermap},
    {"kill",          mhp_kill},
    {"?",             mhp_help},
    {"ldt",           mhp_print_ldt},
+   {"log",           mhp_debuglog},
+   {"dump",          mhp_dump_to_file},
    {"",              NULL}
 };
 
@@ -166,7 +190,14 @@ static const char help_page[]=
   "q                      Quit the debug session\n"
   "kill                   Kill the dosemu process\n"
   "r                      list regs\n"
-  "e ADDR HEXSTR          modify memory (0-1Mb)\n"
+  "e/ed ADDR val [val ..] modify memory (0-1Mb), previous addr for ADDR='-'\n"
+  /* val can be: a hex val (in case of 'e') or decimal (in case of 'ed')
+   * With 'ed' also a hexvalue in form of 0xFF is allowed and can be mixed,
+   * val can also be a character constant (e.g. 'a') or a string ("abcdef"),
+   * val can also be any register symbolic and has the size of that register.
+   * Except for strings and registers, val can be suffixed by
+   * W(word size) or L (long size), default size is byte.
+   */
   "d ADDR SIZE            dump memory (no limit)\n"
   "u ADDR SIZE            unassemble memory (no limit)\n"
   "g                      go (if stopped)\n"
@@ -181,9 +212,13 @@ static const char help_page[]=
   "bpintd/bcintd xx [ax]  set/clear breakpoint on DPMI INT xx [ax]\n"
   "bpload                 stop at start of next loaded DOS program\n"
   "bl                     list active breakpoints\n"
+  "bplog/bclog regex      set/clear breakpoint on logoutput using regex\n"
   "rusermap org fn        read microsoft linker format .MAP file 'fn'\n"
+  "rmapfile [file]        (re)read a dosemu.map ('nm' format) file\n"
   "                       code origin = 'org'.\n"
   "ldt sel lines          dump ldt starting at selector 'sel' for 'lines'\n"
+  "log [flags]            get/set debug-log flags (e.g 'log +M-k')\n"
+  "dump ADDR SIZE FILE    dump a piece of memory to file\n"
   "<ENTER>                repeats previous command\n";
 
 /********/
@@ -354,23 +389,34 @@ static void mhp_rusermap(int argc, char *argv[])
 
 static void mhp_rmapfile(int argc, char *argv[])
 {
+  PRIV_SAVE_AREA
   FILE * ifp;
   unsigned char bytebuf[IBUFS];
   unsigned long a1;
+  char *map_fname = DOSEMU_MAP_PATH;
 
-  ifp = fopen(MAP_LOC, "r");
+  if (argc >= 2) {
+    map_fname = argv[1];
+  }
+  enter_priv_off();
+  ifp = fopen(map_fname, "r");
+  leave_priv_setting();
   if (!ifp) {
-     mhp_printf("unable to open map file\n");
+     mhp_printf("unable to open map file %s\n", map_fname);
      return;
   }
-  mhp_printf("Reading map file\n");
+  mhp_printf("Reading map file %s\n", map_fname);
   last_symbol = 0;
-  for(;;) {
+  while (last_symbol < MAXSYM) {
      if(!fgets(bytebuf, 100, ifp))
         break;
-     if (!strlen(bytebuf))
+     if (!strlen(bytebuf) || !isxdigit(*bytebuf))
         continue;
-     sscanf(&bytebuf[0], "%lx", &a1);
+      /* recent versions of nm put out the address in long long
+       * format, with 16 digits: we can't rely on absolute offsets.
+       */
+      sscanf(bytebuf, "%lx %c %40s", &a1, &symbol_table[last_symbol].type,
+              (char *)&symbol_table[last_symbol].name);
 #ifdef __ELF__
      if (a1 < 0x08000000) continue;
 #else
@@ -380,8 +426,6 @@ static void mhp_rmapfile(int argc, char *argv[])
         break;
 #endif
      symbol_table[last_symbol].addr = a1;
-     symbol_table[last_symbol].type = bytebuf[9];
-     sscanf(&bytebuf[11], "%s", (char *)&symbol_table[last_symbol].name);
      last_symbol++;
   }
   fclose(ifp);
@@ -396,18 +440,28 @@ enum {
   _EAXr,_EBXr,_ECXr,_EDXr,_ESIr,_EDIr,_EBPr,_ESPr,_EIPr
 };
 
+static int last_decode_symreg;
+
 static int decode_symreg(char *regn)
 {
-  static char reg_syms[]="SS  CS  DS  ES  FS  GS EAX EBX ECX EDX ESI EDI EBP ESP EIP  FL";
+  static char reg_syms[]="SS  CS  DS  ES  FS  GS  "
+			 "AX  BX  CX  DX  SI  DI  BP  SP  IP  FL  "
+ 			 "EAX EBX ECX EDX ESI EDI EBP ESP EIP ";
   char rn[5], *s;
   int n;
-  strncpy(rn,regn,4);
-  if (isalpha(rn[2])) rn[3]=0;
-  else rn[2]=0;
-  { char *c=rn; while (*c) *c++ = toupper(*c); }
-  if (!(s=strstr(reg_syms,rn))) return -1;
-  n=s-reg_syms;
-  return (n >> 2) + ((n & 1) ? (16-5) : 0);
+
+  last_decode_symreg = -1;
+  if (!isalpha(*regn))
+	return -1;
+  s=rn; n=0; rn[4]=0;
+  while (n<4)
+  {
+	*s++=(isalpha(*regn)? toupper(*regn++): ' '); n++;
+  }
+  if (!(s = strstr(reg_syms, rn)))
+	return -1;
+  last_decode_symreg = ((s-reg_syms) >> 2);
+  return last_decode_symreg;
 }
 
 static unsigned long mhp_getreg(unsigned char * regn)
@@ -487,7 +541,7 @@ static void mhp_go(int argc, char * argv[])
    } else {
       dpmi_mhp_setTF(0);
       WRITE_FLAGS(READ_FLAGS() & ~TF);
-      vm86s.vm86plus.mhpdbg_TFpendig=0;
+      vm86s.vm86plus.vm86dbg_TFpendig=0;
       mhpdbgc.stopped = 0;
    }
 }
@@ -544,7 +598,7 @@ static void mhp_trace_force(int argc, char * argv[])
         if (!dpmi_mhp_setTF(1)) return;
       }
       else WRITE_FLAGS(READ_FLAGS() | TF);
-      vm86s.vm86plus.mhpdbg_TFpendig=1;
+      vm86s.vm86plus.vm86dbg_TFpendig=1;
       mhpdbgc.trapcmd = 1;
    }
 }
@@ -588,9 +642,10 @@ static void mhp_dis(int argc, char * argv[])
    if (IN_DPMI && seg) data32=dpmi_mhp_get_selector_size(seg);
    for (i=0; i<nbytes; i++) {
       if ((i&0x0f) == 0x00) {
-         if (seg|off)
+         if (seg|off) {
             if (data32) mhp_printf("%s%04x:%08x ", IN_DPMI?"#":"" ,seg, off+i);
             else mhp_printf("%s%04x:%04x ", IN_DPMI?"#":"" ,seg, off+i);
+         }
          else
             mhp_printf( "%08lX ", (unsigned long)seekval+i);
       }
@@ -619,6 +674,50 @@ static void mhp_dis(int argc, char * argv[])
    }
 }
 
+static void mhp_dump_to_file(int argc, char * argv[])
+{
+   PRIV_SAVE_AREA
+   unsigned int nbytes;
+   unsigned long seekval;
+   unsigned char * buf = 0;
+   unsigned int seg;
+   unsigned int off;
+   unsigned int limit=0;
+   int fd;
+
+   if (argc <= 3) {
+      mhp_printf("USAGE: dump <addr> <size> <filesname>\n");
+      return;
+   }
+
+   seekval = (unsigned long)mhp_getadr(argv[1], &seg, &off, &limit);
+
+   buf = (unsigned char *) seekval;
+   sscanf(argv[2], "%x", &nbytes);
+   if (nbytes == 0) {
+      mhp_printf("invalid size\n");
+      return;
+   }
+
+   if (config.secure && ((seekval+nbytes) >0x10ffff)) {
+      mhp_printf("secure is ON, can\'t access above 1MEG\n");
+      return;
+   }
+
+   enter_priv_off();
+   fd = open(argv[3], O_WRONLY | O_CREAT | O_TRUNC, 00775);
+   leave_priv_setting();
+
+   if (fd < 0) {
+      mhp_printf("cannot open/create file %s\n%s\n", argv[3], sys_errlist[errno]);
+      return;
+   }
+   if (write(fd, buf, nbytes) != nbytes) {
+      mhp_printf("write error: %s\n", sys_errlist[errno]);
+   }
+   close(fd);
+}
+
 static void mhp_mode(int argc, char * argv[])
 {
    if (argc >=2) {
@@ -706,7 +805,7 @@ static void mhp_disasm(int argc, char * argv[])
        }
        refseg = seg;
        rc = dis_8086(org+bytesdone, buf+bytesdone, frmtbuf, def_size,
-                  &refseg, &ref, (IN_DPMI ? dpmi_mhp_getselbase(refseg) : 0));
+                  &refseg, &ref, (IN_DPMI ? dpmi_mhp_getselbase(refseg) : 0), 16);
        for (i=0;i<rc;i++) {
            sprintf(&bytebuf[i*2], "%02X", *(buf+bytesdone+i) );
            bytebuf[(i*2)+2] = 0x00;
@@ -737,36 +836,98 @@ static void mhp_disasm(int argc, char * argv[])
       sprintf(lastu, "%lx", seekval + bytesdone);
    }
 }
+
+enum {
+  V_NONE=0, V_BYTE=1, V_WORD=2, V_DWORD=4, V_STRING=8
+};
+
+static int get_value(unsigned long *v, unsigned char *s, int base)
+{
+   int len = strlen(s);
+   int t;
+   char *tt;
+
+   if (!len) return V_NONE;
+   t = s[len-1];
+   if (len >2) {
+     /* check for string value */
+     if (t == '"' && s[0] == '"') {
+       s[len-1] = 0;
+       return V_STRING;
+     }
+   }
+   if ((tt = strchr(" wl", tolower(t))) !=0) {
+     len--;
+     s[len] = 0;
+     t = (int)(tt - " wl") << 1;
+   }
+   else t = V_NONE;
+   if (len >2) {
+     if (s[len-1] == '\'' && s[0] == '\'') {
+       *v = 0;
+       len -=2;
+       if (len > sizeof(*v)) len = sizeof(*v);
+       strncpy((char *)v, s+1, len);
+       if (t != V_NONE) return t;
+       if (len <  2) return V_BYTE;
+       if (len <  4) return V_WORD;
+       return V_DWORD;
+     }
+   }
+   *v = mhp_getreg(s);
+   if (last_decode_symreg >=0) {
+     if (last_decode_symreg < _EAXr) return V_WORD;
+     return V_DWORD;
+   }
+   *v = strtoul(s,0,base);
+   if (t == V_NONE) return V_BYTE;
+   return t;
+}
+
 static void mhp_enter(int argc, char * argv[])
 {
-   int i1;
-   unsigned char * zapaddr;
-   unsigned int seg;
-   unsigned int off;
+   int size;
+   static unsigned char * zapaddr = (char *)-1;
+   unsigned int seg, off;
+   unsigned long val;
    unsigned int limit;
-   unsigned char * inputptr;
-   unsigned char   c;
-   unsigned char   workchrs[4];
+   unsigned char *arg;
+   int base = 16;
 
    if (argc < 3)
       return;
 
-   zapaddr = (unsigned char*)mhp_getadr(argv[1], &seg, &off, &limit);
+   if (!strcmp(argv[1],"-")) {
+     if ((int)zapaddr == -1) {
+        mhp_printf("Address invalid, no previous 'e' command with address\n");
+        return;
+     }
+   }
+   else  zapaddr = (unsigned char*)mhp_getadr(argv[1], &seg, &off, &limit);
    if ((a20 ?0x10fff0 : 0x100000) < (unsigned long)zapaddr) {
       mhp_printf("Address invalid\n");
       return;
    }
-   inputptr = argv[2];
-   for (i1 = 0; ; i1++) {
-      workchrs[0] = *inputptr++;
-      if (!workchrs[0])
-         break;
-      workchrs[1] = *inputptr++;
-      workchrs[2] = 0x00;
-      sscanf(workchrs, "%02x", (unsigned int*)&c);
-      zapaddr[i1] = c;
-      if (!workchrs[1])
-         break;
+   if (!strcmp(argv[0], "ed")) base = 0;
+   argv += 2;
+   while ((arg = *argv) != 0) {
+      size = get_value(&val, arg, base);
+      switch (size) {
+        case V_BYTE:
+        case V_WORD:
+        case V_DWORD: {
+          memcpy(zapaddr, &val, size);
+          zapaddr += size;
+          break;
+        }
+        case V_STRING: {
+          size = strlen(arg+1);
+          memcpy(zapaddr, arg+1, size);
+          zapaddr += size;
+          break;
+        }
+      }
+      argv++;
    }
 }
 
@@ -865,6 +1026,7 @@ static void* mhp_getadr(unsigned char * a1, unsigned int * s1, unsigned int *o1,
    /* Second 32 bits of descriptor */
    base_addr |= (*lp & 0xFF000000) | ((*lp << 16) & 0x00FF0000);
    limit |= (*lp & 0x000F0000);
+   if (*lp & 0x00800000) limit <<= 12;
 
    if ((limit == 0) && (base_addr == 0)) {
      mhp_printf("selector %x appears to be invalid\n", seg1);
@@ -880,7 +1042,6 @@ static void* mhp_getadr(unsigned char * a1, unsigned int * s1, unsigned int *o1,
    return (void *) base_addr + off1;
 }
 
-
 int mhp_setbp(unsigned long seekval)
 {
    int i1;
@@ -915,6 +1076,15 @@ int mhp_clearbp(unsigned long seekval)
    return 0;
 }
 
+static int check_for_stopped(void)
+{
+   if (!mhpdbgc.stopped) {
+     mhp_printf("need to be in 'stopped' state for this command\n");
+     mhp_send();
+   }
+   return mhpdbgc.stopped;
+}
+
 static void mhp_bp(int argc, char * argv[])
 {
    unsigned long seekval;
@@ -922,6 +1092,7 @@ static void mhp_bp(int argc, char * argv[])
    unsigned int off;
    unsigned int limit;
 
+   if (!check_for_stopped()) return;
    if (argc < 2) {
       mhp_printf("location argument required\n");
       return;
@@ -942,7 +1113,7 @@ static void mhp_bl(int argc, char * argv[])
    }
    mhp_printf( "Interrupts: ");
    for (i1=0; i1 < 256; i1++) {
-      if (test_bit(i1, vm86s.vm86plus.mhpdbg_intxxtab)) {
+      if (test_bit(i1, vm86s.vm86plus.vm86dbg_intxxtab)) {
          mhp_printf( "%02x ", i1);
       }
    }
@@ -966,6 +1137,7 @@ static void mhp_bl(int argc, char * argv[])
    }
    mhp_printf( "\n");
    if (mhpdbgc.bpload) mhp_printf("bpload active\n");
+   print_log_breakpoints();
    return;
 }
 
@@ -974,6 +1146,7 @@ static void mhp_bc(int argc, char * argv[])
    int i1;
 
    if (argc <2) return;
+   if (!check_for_stopped()) return;
    if (!sscanf(argv[1], "%d", &i1)) {
      mhp_printf( "Invalid breakpoint number\n");
      return;
@@ -991,12 +1164,13 @@ static void mhp_bpint(int argc, char * argv[])
    int i1;
 
    if (argc <2) return;
+   if (!check_for_stopped()) return; 
    sscanf(argv[1], "%x", &i1);
-   if (test_bit(i1, vm86s.vm86plus.mhpdbg_intxxtab)) {
+   if (test_bit(i1, vm86s.vm86plus.vm86dbg_intxxtab)) {
          mhp_printf( "Duplicate BPINT %02x, nothing done\n", i1);
          return;
    }
-   set_bit(i1, vm86s.vm86plus.mhpdbg_intxxtab);
+   set_bit(i1, vm86s.vm86plus.vm86dbg_intxxtab);
    if (i1 == 0x21) mhpdbgc.int21_count++;
    return;
 }
@@ -1006,12 +1180,13 @@ static void mhp_bcint(int argc, char * argv[])
    int i1;
 
    if (argc <2) return;
+   if (!check_for_stopped()) return; 
    sscanf(argv[1], "%x", &i1);
-   if (!test_bit(i1, vm86s.vm86plus.mhpdbg_intxxtab)) {
+   if (!test_bit(i1, vm86s.vm86plus.vm86dbg_intxxtab)) {
          mhp_printf( "No BPINT %02x, nothing done\n", i1);
          return;
    }
-   clear_bit(i1, vm86s.vm86plus.mhpdbg_intxxtab);
+   clear_bit(i1, vm86s.vm86plus.vm86dbg_intxxtab);
    if (i1 == 0x21) mhpdbgc.int21_count--;
    return;
 }
@@ -1021,6 +1196,7 @@ static void mhp_bpintd(int argc, char * argv[])
    int i1,v1=0;
 
    if (argc <2) return;
+   if (!check_for_stopped()) return; 
    sscanf(argv[1], "%x", &i1);
    i1 &= 0xff;
    if (argc >2) {
@@ -1044,6 +1220,7 @@ static void mhp_bcintd(int argc, char * argv[])
    int i1,v1=0;
 
    if (argc <2) return;
+   if (!check_for_stopped()) return; 
    sscanf(argv[1], "%x", &i1);
    i1 &= 0xff;
    if (argc >2) {
@@ -1069,6 +1246,7 @@ static void mhp_bcintd(int argc, char * argv[])
 
 static void mhp_bpload(int argc, char * argv[])
 {
+   if (!check_for_stopped()) return; 
    if (mhpdbgc.bpload) {
      mhp_printf("load breakpoint already pending\n");
      return;
@@ -1076,35 +1254,12 @@ static void mhp_bpload(int argc, char * argv[])
    mhpdbgc.bpload=1;
    {
      volatile register int i=0x21; /* beware, set_bit-macro has wrong constraints */
-     set_bit(i, vm86s.vm86plus.mhpdbg_intxxtab);
+     set_bit(i, vm86s.vm86plus.vm86dbg_intxxtab);
    }
    mhpdbgc.int21_count++;
    return;
 }
 
-static int mhp_parse(char *s, char* argvx[])
-{
-   int mode = 0;
-   int argcx = 0;
-
-   for ( ; *s; s++) {
-      if (!mode) {
-         if (*s > ' ') {
-            mode = 1;
-            argvx[argcx++] = s;
-            if (argcx >= MAXARG)
-               break;
-         }
-      } else {
-         if (*s <= ' ') {
-            mode = 0;
-            *s = 0x00;
-         }
-      }
-   }
-   return(argcx);
-}
-
 static void mhp_regs(int argc, char * argv[])
 {
   unsigned long newval;
@@ -1145,7 +1300,8 @@ static void mhp_regs(int argc, char * argv[])
      mhp_printf( "General Protection Fault, ");
 
  if (!traceloop)
-  mhp_printf( "system state: %s%s%s\n",
+  mhp_printf( "system state: %s%s%s%s\n",
+       "",
        mhpdbgc.stopped ? "stopped" : "running",
        IN_DPMI ? " in DPMI" : (in_dpmi?" in real mode while in DPMI":""),
        IN_DPMI ?(dpmi_mhp_getcsdefault()?"-32bit":"-16bit") : "");
@@ -1187,7 +1343,6 @@ static void mhp_regs32(int argc, char * argv[])
 
 static void mhp_kill(int argc, char * argv[])
 {
-  extern void mhp_send(void), mhp_close(void), mhp_putc(char c1);
   mhp_cmd("r0");
   mhp_printf("\ndosemu killed via debug terminal\n");
   mhp_send();
@@ -1196,7 +1351,8 @@ static void mhp_kill(int argc, char * argv[])
   mhp_send();
 #endif
   mhp_close();
-  leavedos(1);
+  if (dosdebug_flags & DBGF_IN_LEAVEDOS) dosdebug_flags &= ~DBGF_IN_LEAVEDOS;
+  else leavedos(1);
 }
 
 static void mhp_help(int argc, char * argv[])
@@ -1270,7 +1426,7 @@ int mhp_getcsip_value()
 
 void mhp_modify_eip(int delta)
 {
-  extern dpmi_mhp_modify_eip(int delta);
+  extern int dpmi_mhp_modify_eip(int delta);
   if (IN_DPMI) dpmi_mhp_modify_eip(delta);
   else LWORD(eip) +=delta;
 }
@@ -1278,27 +1434,7 @@ void mhp_modify_eip(int delta)
 
 void mhp_cmd(const char * cmd)
 {
-   int argc1;
-   char * argv1[MAXARG];
-   char tmpcmd[250];
-   void (*cmdproc)(int, char *[]);
-   const struct cmd_db * cmdp;
-
-   strcpy(tmpcmd, cmd);
-   argc1 = mhp_parse(tmpcmd, argv1);
-   if (argc1 < 1)
-      return;
-   for (cmdp = cmdtab, cmdproc = NULL; cmdp->cmdproc; cmdp++) {
-      if (!memcmp(cmdp->cmdname, argv1[0], strlen(argv1[0])+1)) {
-         cmdproc = cmdp->cmdproc;
-         break;
-      }
-   }
-   if (!cmdproc) {
-      mhp_printf("Command %s not found\n", argv1[0]);
-      return;
-   }
-   (*cmdproc)(argc1, argv1);
+   call_cmd(cmd, MAXARG, cmdtab, mhp_printf);
 }
 
 static void mhp_print_ldt(int argc, char * argv[])
@@ -1385,3 +1521,203 @@ static void mhp_print_ldt(int argc, char * argv[])
   sprintf (lastldt, "%x", i);
 }
 
+static void mhp_debuglog(int argc, char * argv[])
+{
+   char buf[256];
+   if (argc >1) {
+     if (!strcmp(argv[1], "on")) {
+       dosdebug_flags |= DBGF_INTERCEPT_LOG | DBGF_LOG_TO_DOSDEBUG;
+       return;
+     }
+     if (!strcmp(argv[1], "off")) {
+       dosdebug_flags &= ~DBGF_LOG_TO_DOSDEBUG;
+       if (!(dosdebug_flags & DBGF_LOG_TO_BREAK))
+         dosdebug_flags &= ~DBGF_INTERCEPT_LOG;
+       return;
+     }
+     SetDebugFlagsHelper(argv[1]);
+   }
+   if (argc <=1) {
+     GetDebugFlagsHelper(buf, 0);
+     mhp_printf ("current Debug-log flags:\n%s\n", buf);
+   }
+}
+
+#define MAX_REGEX		8
+static regex_t *rxbuf[MAX_REGEX] = {0};
+static char *rxpatterns[MAX_REGEX] = {0};
+static int num_regex = 0;
+static int num_logbp = 0;
+
+static int get_free_regex_buf(void)
+{
+  int i;
+  for (i=0; i <num_regex; i++) {
+    if (!rxbuf[i]) return i;
+  }
+  if (num_regex >= MAX_REGEX) return -1;
+  i = num_regex;
+  num_regex++;
+  rxbuf[i] = 0;
+  rxpatterns[i] =0;
+  return i;
+}
+
+static void free_regex(int number)
+{
+  if ((unsigned)number >= MAX_REGEX) return;
+  if (!rxbuf[number]) return;
+  regfree(rxbuf[number]);
+  if (rxpatterns[number]) free(rxpatterns[number]);
+  free(rxbuf[number]);
+  rxbuf[number] = 0;
+  rxpatterns[number] = 0;
+}
+
+static char *trimm_string_arg(char *arg)
+{
+  int len;
+
+  if (!arg || !arg[0]) return 0;
+  len = strlen(arg);
+  if (     (arg[0] == '"' && arg[len-1] == '"')
+        || (arg[0] == '\'' && arg[len-1] == '\'')) {
+    arg[len-1] = 0;
+    return arg+1;
+  }
+  return arg;
+}
+
+static void print_log_breakpoints(void)
+{
+   int rx;
+
+   num_logbp = 0;
+   for (rx=0; rx <num_regex; rx++) {
+     if (rxbuf[rx]) {
+       mhp_printf("log break point %d: %s\n", rx, rxpatterns[rx]);
+       mhp_send();
+       num_logbp++;
+     }
+   }
+   if (!num_logbp) {
+     mhp_printf("no log break points aktiv\n");
+   }
+}
+
+
+static void mhp_bplog(int argc, char * argv[])
+{
+   int rx, errcode;
+   char buf[1024];
+   char *s;
+
+   if (argc >1) {
+     if (!check_for_stopped()) return; 
+     argv++;
+     buf[0] = 0;
+     while (*argv) {
+       s = trimm_string_arg(*argv);
+       if (s) strcat(buf, s);
+       argv++;
+     }
+     if (!buf[0]) {
+       mhp_printf ("no valid regular expression defined\n");
+       return;
+     }
+     rx = get_free_regex_buf();
+     if (rx <0) {
+       mhp_printf ("to many log breakpoints, use bclog to free one\n");
+       return;
+     }
+     rxpatterns[rx] = strdup(buf);
+     rxbuf[rx] = malloc(sizeof(regex_t));
+     if (!rxbuf[rx]) {
+       mhp_printf ("out of memory\n");
+       return;
+     }
+     errcode = regcomp(rxbuf[rx], rxpatterns[rx], REG_NOSUB);
+     if (errcode) {
+        regerror(errcode, rxbuf[rx], buf, sizeof(buf));
+        mhp_printf ("%s\n", buf);
+        free_regex(rx);
+        return;
+     }
+     /* ...puh, all that work just for generating a pattern :-)
+      * now we enable the 'break point' by intercepting the log
+      */
+     dosdebug_flags |= DBGF_INTERCEPT_LOG | DBGF_LOG_TO_BREAK;
+   }
+   /* atleast print all active brakepoints */
+   print_log_breakpoints();
+}
+
+static void mhp_bclog(int argc, char * argv[])
+{
+   int rx;
+
+   if (argc >1) {
+     if (!check_for_stopped()) return;
+     rx = atoi(argv[1]);
+     if (((unsigned)rx >= MAX_REGEX) || !rxbuf[rx]) {
+       mhp_printf("log break point not existing\n", rx);
+       return;
+     }
+     free_regex(rx);
+   }
+   print_log_breakpoints();
+   if (!num_logbp) {
+     dosdebug_flags &= ~DBGF_LOG_TO_BREAK;
+     if (!(dosdebug_flags & DBGF_LOG_TO_DOSDEBUG)) {
+       dosdebug_flags &= DBGF_INTERCEPT_LOG;
+     }
+   }
+}
+
+static int mhp_check_regex(char *line)
+{
+   int rx, hit;
+   if (!num_logbp) return 0;  /* we should not come here */
+   for (rx=0; rx <num_regex; rx++) {
+     if (rxbuf[rx]) {
+       hit = regexec(rxbuf[rx], line, 0, 0, 0) == 0;
+       if (hit) {
+         mhp_printf("log break point %d hit: >%s<\n",rx,line);
+         mhp_send();
+       }
+       if (hit) return 1;
+     }
+   }
+   return 0;
+}
+
+static char lbuf[1024];
+static int lbufi = 0;
+
+void mhp_regex(const char *fmt, va_list args)
+{
+  int i, hit;
+  char *s;
+  
+  if (!(dosdebug_flags & DBGF_LOG_TO_BREAK)) return;
+  lbufi += vsprintf(lbuf+lbufi, fmt, args);
+  hit = 0;
+  i = 0;
+  do {
+    s = strchr(lbuf+i, '\n');
+    if (s) {
+      *s = 0;
+      hit = mhp_check_regex(lbuf+i);
+      i = (s - lbuf)+1;
+      if (hit) break;
+    }
+  } while (s);
+  if (i) {
+    memcpy(lbuf, lbuf+i, lbufi-i+1);
+    lbufi -= i;
+    if (hit) {
+      mhpdbgc.want_to_stop = 1;
+    }
+  }
+}
+
