@@ -1,13 +1,46 @@
-/* this handles the int10 video functions.
-   Most functions here change only the video memory and status 
-   variables; the actual screen is then rendered asynchronously 
-   after these by Video->update_screen.
+/* 
+ * (C) Copyright 1992, ..., 1999 the "DOSEMU-Development-Team".
+ *
+ * for details see file COPYING in the DOSEMU distribution
+ */
 
-  5/24/95, Erik Mouw (J.A.K.Mouw@et.tudelft.nl) and 
-  Arjan Filius (I.A.Filius@et.tudelft.nl)
-  changed int10() to make graphics work with X.
-*/
+/* 
+ * DANG_BEGIN_MODULE
+ *
+ * REMARK
+ * Video BIOS implementation.
+ *
+ * This module handles the int10 video functions.
+ * Most functions here change only the video memory and status 
+ * variables; the actual screen is then rendered asynchronously 
+ * after these by Video->update_screen.
+ *
+ * /REMARK
+ * DANG_END_MODULE
+ *
+ * DANG_BEGIN_CHANGELOG
+ *
+ * 5/24/95, Erik Mouw (J.A.K.Mouw@et.tudelft.nl) and 
+ * Arjan Filius (I.A.Filius@et.tudelft.nl)
+ * changed int10() to make graphics work with X.
+ *
+ * 1998/04/05: Put some work into set_video_mode() (made it
+ * more VGA comaptible) and removed new_set_video_mode().
+ * Removed (useless) global variable "gfx_mode".
+ * -- sw (Steffen Winterfeldt <wfeldt@suse.de>)
+ * 
+ * Readded new_set_video_mode, its needed for non-X compiles.
+ * -- EB 3 June 1998
+ *
+ * Renamed new_set_video_mode to X_set_video_mode to avoid confusion
+ * in the future
+ * -- Hans 980614
+ *
+ * DANG_END_CHANGELOG
+ */
 
+
+#include <string.h>
 #include "config.h"
 
 #include "emu.h"
@@ -15,13 +48,11 @@
 #include "memory.h"
 #include "bios.h"
 #include "vc.h"
+#include "speaker.h"
 
 #if X_GRAPHICS
 #include "X.h"
 #include "vgaemu.h"
-#ifdef VESA /* root@zaphod */
-#include "vesa.h"
-#endif
 #endif
 
 /* a bit brutal, but enough for now */
@@ -59,6 +90,9 @@
 inline void set_cursor_shape(ushort shape) {
    int cs,ce;
 
+   if (config.cardtype == CARD_NONE)
+     return;
+
    cs=CURSOR_START(shape) & 0x1F;
    ce=CURSOR_END(shape) & 0x1F;
 
@@ -102,6 +136,9 @@ Scroll(us *sadr, int x0, int y0, int x1, int y1, int l, int att)
   us blank = ' ' | (att << 8);
   us tbuf[MAX_COLUMNS];
 
+  if (config.cardtype == CARD_NONE)
+     return;
+
   if ( (config.cardtype == CARD_MDA) && ((att & 7) != 0) && ((att & 7) != 7) )
     {
       blank = ' ' | ((att | 7) << 8);
@@ -190,6 +227,9 @@ void bios_scroll(int x0,int y0,int x1,int y1,int n,byte attr) {
    struct scroll_entry *s;
    int sqh2;
 
+   if (config.cardtype == CARD_NONE)
+     return;
+
    VIDEO_UPDATE_LOCK();
    sqh2=SQ_INC(sq_head);
    if (n!=0 && !Video->is_mapped && sqh2!=sq_tail) {
@@ -229,9 +269,16 @@ char_out(unsigned char ch, int s)
 {
   int newline_att = 7;
   int xpos, ypos;
+
 #if USE_DUALMON
   int virt_text_base = BIOS_SCREEN_BASE;
 #endif
+
+  if (config.cardtype == CARD_NONE) {
+     putchar (ch);
+     return;
+  }
+
   xpos = get_bios_cursor_x_position(s);
   ypos = get_bios_cursor_y_position(s);
 
@@ -302,6 +349,9 @@ clear_screen(int s, int att)
   int virt_text_base = BIOS_SCREEN_BASE;
 #endif
 
+  if (config.cardtype == CARD_NONE)
+     return;
+
   v_printf("VID: cleared screen: %d %d %p\n", s, att, SCREEN_ADR(s));
   if (s > max_page) return;
   
@@ -318,6 +368,145 @@ clear_screen(int s, int att)
 }
 
 
+#if X_GRAPHICS
+
+/*
+ * set_video_mode() accepts both (S)VGA and VESA mode numbers
+ * It should be fully compatible with the old set_video_mode()
+ * function.
+ * Note: bit 16 in "mode" is used internally to indicate that
+ * nothing should be done except adjusting the font size.
+ * -- 1998/04/04 sw
+ */
+
+static boolean X_set_video_mode(int mode) {
+  vga_mode_info *vmi;
+  int clear_mem = 1;
+  int adjust_font_size = 0;
+  
+  if (config.cardtype == CARD_NONE) {
+    v_printf("set_video_mode: no video!\n");
+    return 0;
+  }
+
+
+  if(mode & (1 << 16)) {
+    adjust_font_size = 1;
+    mode &= ~(1 << 16);
+  }
+
+  mode &= 0xffff;
+
+  if(mode != video_mode) adjust_font_size = 0;
+
+  v_printf("set_video_mode: mode = 0x%02x%s\n", mode, adjust_font_size ? " (adjust font size)" : "");
+
+  if((vmi = vga_emu_find_mode(mode, NULL)) == NULL) {
+    v_printf("set_video_mode: undefined video mode\n");
+    return 0;
+  }
+
+  if(Video->setmode == NULL) {
+    v_printf("set_video_mode: no setmode handler!\n");
+    return 0;
+  }
+
+  if(adjust_font_size) {
+    text_scanlines = vmi->height;
+    li = text_scanlines / vga_font_height;
+    if(li > MAX_LINES) li = MAX_LINES;
+    WRITE_BYTE(BIOS_ROWS_ON_SCREEN_MINUS_1, li - 1);
+    WRITE_WORD(BIOS_FONT_HEIGHT, vga_font_height);
+    Video->setmode(vmi->mode_class, co, li);
+    return 1;
+  }
+
+  video_mode = mode;
+
+  if(mode >= 0x80 && mode < 0x100) {
+    mode &= 0x7f;
+    clear_mem = 0;
+  }
+  if(mode & 0x8000) {
+    mode &= ~0x8000;
+    clear_mem = 0;
+  }
+
+  if(config.cardtype == CARD_MDA) video_mode = 7;
+
+  /*
+   * We store the SVGA mode number (if possible) even when setting a VESA mode. -- sw
+   */
+  WRITE_BYTE(BIOS_VIDEO_MODE, vmi->mode & 0x7f);
+  WRITE_BYTE(BIOS_CURRENT_SCREEN_PAGE, 0);
+
+  if(vmi->mode_class == TEXT) {
+    clear_scroll_queue();
+  }
+
+  li = vmi->text_height;
+  co = vmi->text_width;
+  if(li > MAX_LINES) li = MAX_LINES;
+  vga_font_height = vmi->char_height;
+  text_scanlines = vmi->height;
+
+  WRITE_BYTE(BIOS_ROWS_ON_SCREEN_MINUS_1, li - 1);
+  WRITE_WORD(BIOS_SCREEN_COLUMNS, co);
+  WRITE_WORD(BIOS_FONT_HEIGHT, vga_font_height);
+
+#if USE_DUALMON
+  /*
+   * The following code (& comment) is taken literally
+   * from the old set_video_mode(). Don't know what it's
+   * for (or if it works). -- 1998/04/04 sw 
+   */
+
+  /*
+   * This to be sure in case of older DOS programs probing HGC.
+   * There was no secure way to detect a HGC before VGA was invented.
+   * ( Now we can do INT 10, AX=1A00 ).
+   * Some older DOS programs do it by modifying EQUIP-flags
+   * and then let the BIOS say, if it can ?!?!)
+   * If we have config.dualmon, this happens legaly.
+   */
+  if(config.dualmon)
+    Video->setmode(7, co, li);
+  else
+#endif
+
+  /* setmode needs video_mode to _still have_ the memory-clear bit -- sw */
+  Video->setmode(vmi->mode_class, co, li);
+
+  /*
+   * video_mode is expected to be the mode number _without_ the
+   * memory-clear bit
+   * -- sw
+   */
+  video_mode = mode;
+
+  if(clear_mem && vmi->mode_class == TEXT) clear_screen(0, 7);
+
+  WRITE_BYTE(BIOS_VIDEO_INFO_0, clear_mem ? 0x60 : 0xe0);
+  memset((void *) 0x450, 0, 0x10);	/* equiv. to set_bios_cursor_(x/y)_position(0..7, 0) */
+
+  video_page = 0;
+  WRITE_WORD(BIOS_VIDEO_MEMORY_ADDRESS, 0);
+  screen_adr = SCREEN_ADR(0);
+  screen_mask = 0;
+  cursor_col = get_bios_cursor_x_position(0);
+  cursor_row = get_bios_cursor_y_position(0);
+
+  /*
+   * There are still some BIOS variables that need to be updated.
+   * (cursor shape, CRTC address...)
+   * -- sw
+   */
+
+  return 1;
+}    
+#endif /* X_GRAPHICS */
+
+
 /* XXX- shouldn't this reset the current video page to 0 ? 
 */
 
@@ -327,6 +516,9 @@ boolean set_video_mode(int mode) {
 
   v_printf("set_video_mode: mode = 0x%02x\n",mode);
   
+#ifdef X_GRAPHICS
+  if (Video == &Video_X) return X_set_video_mode(mode);
+#endif  
   oldco = co;
   old_video_mode = video_mode;
   video_mode=mode&0x7f;
@@ -389,7 +581,6 @@ boolean set_video_mode(int mode) {
 #endif
     
 do_text_mode:
-    gfx_mode = TEXT;	        /* we're in a text mode now */
     clear_scroll_queue();
 
     li=text_scanlines/vga_font_height;
@@ -405,15 +596,16 @@ do_text_mode:
     if ( config.cardtype == CARD_MDA )
       mode = 7;
        
-    WRITE_BYTE(BIOS_VIDEO_MODE, video_mode=mode&0x7f);
     break;
 
+
 case 0x13:	/*Not finished ! */
 case 0x5c:
 case 0x5d:
 case 0x5e:
 case 0x62:
   /* 0x01 == GRAPH for us, but now it's sure! */
+  if (Video->setmode != NULL)
   Video->setmode(0x01,0,0);
   break;
 
@@ -422,6 +614,9 @@ case 0x62:
     v_printf("undefined video mode 0x%x\n", mode);
   goto error;
   }
+
+  WRITE_BYTE(BIOS_VIDEO_MODE, video_mode=mode&0x7f);
+
   if (oldco != co)
     WRITE_WORD(BIOS_SCREEN_COLUMNS, co);
   return 1;
@@ -432,6 +627,50 @@ error:
   return 0;
 }    
 
+/* get the active and alternate display combination code */
+void get_dcc(int *active_dcc, int *alternate_dcc)
+{
+#if USE_DUALMON
+  if (config.dualmon) {
+    if (IS_SCREENMODE_MDA) {
+      *active_dcc = MDA_VIDEO_COMBO;  /* active display */
+      *alternate_dcc = video_combo;
+    }
+    else {
+      *active_dcc = video_combo;     /* active display */
+      *alternate_dcc = MDA_VIDEO_COMBO;
+    }
+    return;
+  }
+#endif
+  *active_dcc = video_combo;	/* active display */
+  *alternate_dcc = 0;		/* no inactive display */
+}
+
+/* INT 10 AH=1B - FUNCTIONALITY/STATE INFORMATION (PS,VGA/MCGA) */
+void return_state(Bit8u *statebuf) {
+	int active_dcc, alternate_dcc;
+	
+	memset(statebuf, 0, 4); /* XXX pointer to static functionality table */
+	/* store bios 0:449-0:466 at ofs 0x04 */
+	memcpy(statebuf + 0x04, (char *)0x449, 0x466 - 0x449 + 1);
+	/* store bios 0:484-0:486 at ofs 0x22 */
+	memcpy(statebuf + 0x22, (char *)0x484, 0x486 - 0x484 + 1);
+	get_dcc(&active_dcc, &alternate_dcc);
+	statebuf[0x25] = active_dcc;
+	statebuf[0x26] = alternate_dcc;
+	statebuf[0x27] = 16;/* XXX number of colors, low byte */
+	statebuf[0x28] = 0; /* XXX number of colors, high byte */
+	statebuf[0x29] = 8; /* XXX number of pages supported */
+	statebuf[0x2a] = 2; /* XXX number of scanlines 0-3=200,350,400,480 */
+	statebuf[0x2b] = 0; /* XXX primary character block */
+	statebuf[0x2c] = 0; /* XXX secondary character block */
+	statebuf[0x31] = 3; /* video memory: 3 = 256K */
+	statebuf[0x32] = 0; /* XXX save pointer state flags */
+	memset(statebuf + 0x33, 0, 13);
+}
+
+
 /******************************************************************/
 
 /* the actual int10 handler */
@@ -482,7 +721,7 @@ void int10()
     /* set appropriate font height for 25 lines */
     vga_font_height=text_scanlines/25;
     if (!set_video_mode(LO(ax))) {
-       error("int10,0: set_video_mode failed\n");
+       v_printf("int10,0: set_video_mode failed\n");
        CARRY;
     }
     break;
@@ -497,13 +736,19 @@ void int10()
     page = HI(bx);
     x = LO(dx);
     y = HI(dx);
+    v_printf("set cursor: pg:%d x:%d y:%d\n", page, x, y);
     if (page > 7) {
-      error("ERROR: video error (setcur/page>7: %d)\n", page);
+      v_printf("ERROR: video error (setcur/page>7: %d)\n", page);
       CARRY;
       return;
     }
-    if (x >= co || y >= li)
-      break;
+    if (x >= co || y >= li) {
+      /* some apps use this to hide the cursor,
+       * we move it 1 char behind the visible part
+       */
+      x = co;
+      y = li -1;
+    }
 
     set_bios_cursor_x_position(page, x);
     set_bios_cursor_y_position(page, y);
@@ -514,7 +759,7 @@ void int10()
   case 0x3:			/* get cursor pos/shape */
     page = HI(bx);
     if (page > 7) {
-      error("ERROR: video error(0x3 page>7: %d)\n", page);
+      v_printf("ERROR: video error(0x3 page>7: %d)\n", page);
       CARRY;
       return;
     }
@@ -531,13 +776,13 @@ void int10()
       page = LO(ax);
       v_printf("VID: change page from %d to %d!\n", READ_BYTE(BIOS_CURRENT_SCREEN_PAGE), page);
       if (page > max_page) {
-	error("ERROR: video error: set bad page %d\n", page);
+	v_printf("ERROR: video error: set bad page %d\n", page);
 	CARRY;
 	break;
       }
       if (config.console_video) set_vc_screen_page(page);
 #if X_GRAPHICS
-      if (config.X) set_vgaemu_page(page);      /*root@sjoerd*/
+      if (config.X) vga_emu_set_text_page(page, TEXT_SIZE);
 #endif
 
       WRITE_BYTE(BIOS_CURRENT_SCREEN_PAGE, video_page = page);
@@ -563,7 +808,7 @@ void int10()
   case 0x8:			/* read character at x,y + attr */
     page = HI(bx);
     if (page > max_page) {
-      error("ERROR: read char from bad page %d\n", page);
+      v_printf("ERROR: read char from bad page %d\n", page);
       CARRY;
       break;
     }
@@ -627,9 +872,47 @@ void int10()
       
       if (LO(ax) < 2) {                  /* use attribute in BL */
         attr=LO(bx)<<8;
+#if 1					/* 1998/03/13 sengoku@intsys.co.jp */
+        x += n;
+	while(n--) {
+	  /* *(sadr++) = *src++ | attr; */
+          if (*src <= 0x1a) {
+            switch(*src) {
+              case 0x0d :
+                sadr = (u_short *) SCREEN_ADR(page) + y*co, x = n;
+                src++;
+                break;
+              case 0x0a :
+                sadr += co, x--, y++;
+                src++;
+                break;
+              case 0x07 : /* beep */
+#if 0  /* ignored, we have to check wether it breaks the speaker stuff --Hans */
+                putchar(0x07);
+#else
+                /* this should do what we want, says Eric 980425 */
+                speaker_on(125, 0x637);
+#endif
+                x--;
+                src++;
+                break;
+              case 0x08 : /* backspace */
+                sadr--, x--;
+                src++;
+                break;
+              default:
+                WRITE_WORD(sadr++, READ_BYTE(src++) | attr);
+                break;
+            }
+          } else {
+            WRITE_WORD(sadr++, READ_BYTE(src++) | attr);
+          }
+        }
+#else
 	while(n--)
 	  /* *(sadr++) = *src++ | attr; */
-	  WRITE_WORD(sadr++, READ_BYTE(src++) | attr);
+            WRITE_WORD(sadr++, READ_BYTE(src++) | attr);
+#endif
       }
       else {                             /* use attributes in buffer */
 	/*memcpy(sadr,src,n*2);*/
@@ -672,11 +955,40 @@ void int10()
     if(config.X)
       {
         int i, count;
-        unsigned char* src;
+        unsigned char *src;
         unsigned char r, g, b, index;
+	DAC_entry rgb;
 
         switch(LO(ax))
           {
+          case 0x00:
+            Attr_set_entry(LO(bx), HI(bx));
+            break;
+
+          case 0x01:
+            Attr_set_entry(0x11, HI(bx));
+            break;
+
+          case 0x02:
+            src = SEG_ADR((unsigned char *), es, dx);
+            for(i = 0; i < 0x10; i++) Attr_set_entry(i, src[i]);
+            Attr_set_entry(0x11, src[i]);
+            break;
+
+          case 0x07:
+            HI(bx) = Attr_get_entry(LO(bx));
+            break;
+
+          case 0x08:
+            HI(bx) = Attr_get_entry(0x11);
+            break;
+
+          case 0x09:
+            src = SEG_ADR((unsigned char *), es, dx);
+            for(i = 0; i < 0x10; i++) src[i] = Attr_get_entry(i);
+            src[i] = Attr_get_entry(0x11);
+            break;
+
           case 0x10:
             DAC_set_entry((unsigned char)HI(dx), (unsigned char)HI(cx),
                           (unsigned char)LO(cx), (unsigned char)LO(bx));
@@ -695,6 +1007,26 @@ void int10()
               }
             break;
 
+          case 0x15:  /* Read Individual DAC Register */
+            DAC_read_entry(&rgb, (unsigned char)LO(bx));
+            (unsigned char)HI(dx) = rgb.r;
+            (unsigned char)HI(cx) = rgb.g;
+            (unsigned char)LO(cx) = rgb.b;
+            break;
+
+          case 0x17:  /* Read Block of DAC Registers */
+            index=(unsigned char)LO(bx);
+            count=LWORD(ecx);
+            src=SEG_ADR((unsigned char*),es,dx);
+            for(i=0; i<count; i++, index++)
+              {
+                DAC_read_entry(&rgb, index); 
+                src[i*3]=rgb.r;
+                src[i*3+1]=rgb.g;
+                src[i*3+2]=rgb.b;
+              }
+            break;
+
           default:
             break;
           }
@@ -710,7 +1042,7 @@ void int10()
 #endif
 
   case 0x4:			/* get light pen */
-    error("ERROR: video error(no light pen)\n");
+    v_printf("ERROR: video error(no light pen)\n");
 #if 0
     CARRY;
 #else
@@ -721,23 +1053,12 @@ void int10()
 
   case 0x1a:			/* get display combo */
     if (LO(ax) == 0) {
+      int active_dcc, alternate_dcc;
       v_printf("get display combo!\n");
       LO(ax) = 0x1a;		/* valid function=0x1a */
-#if USE_DUALMON
-      if (config.dualmon) {
-        if (IS_SCREENMODE_MDA) {
-          LO(bx) = MDA_VIDEO_COMBO;  /* active display */
-          HI(bx) = video_combo;
-        }
-        else {
-          LO(bx) = video_combo;     /* active display */
-          HI(bx) = MDA_VIDEO_COMBO;
-        }
-        break;
-      }
-#endif
-      LO(bx) = video_combo;	/* active display */
-      HI(bx) = 0;		/* no inactive display */
+      get_dcc(&active_dcc, &alternate_dcc);
+      LO(bx) = active_dcc;
+      HI(bx) = alternate_dcc;
     }
     else {
       v_printf("set display combo not supported\n");
@@ -778,7 +1099,7 @@ void int10()
 	
       more_lines:
 	{
-	  if (!set_video_mode(video_mode | 0x80)) 
+	  if(!set_video_mode(video_mode | (1 << 16))) 
 	    {
 	      li = old_li; /* not that it changed */
 	      vga_font_height = old_font_height;
@@ -849,27 +1170,36 @@ void int10()
       break;
       
     default:
-      error("ERROR: unrecognized video subsys config!!\n");
+      v_printf("ERROR: unrecognized video subsys config!!\n");
       show_regs(__FILE__, __LINE__);
     }
     break;
     
+  case 0xcc:			/* called from NC 5.0 */
+    _CX = 0; _AL = 0xff;
+    break;
+
   case 0xfe:			/* get shadow buffer..return unchanged */
   case 0xff:			/* update shadow buffer...do nothing */
     break;
 
+  case 0x1b:                    /* functionality/state information */
+    if (LWORD(ebx) == 0) {
+      v_printf("return functionality/state information");
+      return_state(SEG_ADR((Bit8u *), es, di));
+      LO(ax) = 0x1b;
+    } else
+      v_printf("unknown functionality/state request: %d", LWORD(ebx));
+    break;
 #if 0
-  case 0x1b:                    /* return state */
   case 0x1c:                    /* return save/restore */
 #endif
 
 #if X_GRAPHICS
-#ifdef VESA /* root@zaphod */
   case 0x4f:                    /* vesa interrupt */
     if(config.X)
       do_vesa_int();
     break;
-#endif
 #endif
 
   default:
