@@ -1,9 +1,17 @@
+/* 
+ * (C) Copyright 1992, ..., 1999 the "DOSEMU-Development-Team".
+ *
+ * for details see file COPYING in the DOSEMU distribution
+ */
+
 /*  DANG_BEGIN_MODULE
  *
+ * REMARK
  *  pic.c is a fairly complete emulation of both 8259 Priority Interrupt 
  *  Controllers.  It also includes provision for 16 lower level interrupts.
  *  This implementation supports the following i/o commands:
  *
+ * VERB
  *      ICW1    bits 0 and 1     number of ICWs to expect
  *      ICW2    bits 3 - 7       base address of IRQs
  *      ICW3    no bits          accepted but ignored
@@ -15,6 +23,7 @@
  *                               select special mask mode
  *
  *     Reads of both pic ports are supported completely.
+ * /VERB
  *
  *  An important concept to understand in pic is the interrupt level.
  *  This is a value which represents the priority of the current interrupt.  
@@ -37,9 +46,87 @@
  *  or 0r will turn off debugging messages.  Flags may be combined:
  *  to get both 1r and 2r, use a value of 3r.
  *
+ * /REMARK
  * DANG_END_MODULE
  */
 
+/*
+ * > 
+ * > In 0.66.x the new interrupt is scheduled too early, and restore_rm_regs()
+ * > is never called. Then the crash.
+ * > 
+ * > There are at least 2 problems with this kind of code:
+ * > 1) avoiding interrupt reentrancy
+ * > 2) avoiding too early interrupt scheduling
+ * > 
+ *
+ * You have run into an old problem, which was created through attempts
+ * (not by me) to speed up pic by creating holes in its re-entrancy protection.
+ * As I released pic, it was quite safe on this score, however, that made
+ * dosemu unpleasantly slow.  For everyone's edification, here's the problem:
+ * 
+ * As dos is designed, it is sometimes necessary to presume that an interrupt
+ * (most often the timer, but sometimes the keyboard or even a serial port)
+ * could not possibly re-trigger before a certain amount of time has elapsed.
+ * Sloppy dos programmers have also often made this assumption even when it was
+ * not necessary.
+ *
+ * Dosemu breaks this assumption, because dos doesn't always have the entire
+ * machine, and interrupts can "pile-up" while other processes are running.
+ * As far as pic (hardware or emulated) is concerned, an interrupt ends
+ * when the interrupt acknowledge is sent to pic.  However, to avoid
+ * re-entrancy, pic.c points the iret to a HLT, which is trapped, and used to
+ * allow re-triggerring of an interrupt.  This works, and would completely
+ * eliminate the above problem, but there's more to the story.
+ *
+ * Since some programs put great piles of code inside an interrupt ("I've ACK'd
+ * the interrupt, what's the hurry on the IRET?"), this kind of hold-off slows
+ * dosemu down.  A few (2 or 3) people have modified pic since I released it
+ * to create holes in the hold-off mechanism, citing the need for more speed.
+ * They were certainly right that speed was needed, but I could never get them
+ * to understand the problem they were creating.  These changes are what's
+ * creating the re-entrancy, I believe.
+ *
+ * To make matters worse, dos uses IRETs for software interrupts, even inside
+ * hardware interrupt routines.  So the only reliable way to catch the IRET of
+ * a hardware interrupt is to point it to a HLT by modifying the stack.  Please
+ * don't question this, I've many hours of debugging time that demonstrated
+ * this quite thoroughly.
+ *
+ * BUT...task-switching environments (MS-Win) use hardware interrupts (mostly
+ * timer and keyboard, but sometimes serial, too) to switch tasks. When a new
+ * task is created, there is no address pointing to a HLT on the (new) stack,
+ * so the end of the interrupt code will never be sensed, and the re-trigger
+ * protection will lock up dosemu.  So it is necessary to detect this event,
+ * and decide that the interrupt has terminated, or we'll never let the timer
+ * (or keyboard, etc. as the case may be) interrupt re-trigger.  So, in a way,
+ * we MUST create some kind of hole in the hold-off process.
+ *
+ * The entire pic debate is one of finding the right kind of hole to make, and
+ * the trade-off is robustness vs. speed. Since the task switch problem only
+ * occurs on the initial invocation of a dos task (when a new stack is built),
+ * I elected to use a conservative hole, to maintain robustness.  Others have
+ * since modified pic.c to gain speed, often without adequate testing of
+ * robustness.
+ * 
+ * There's lots of re-entrancy protection in pic, but it has largely been
+ * defeated by speed-up efforts.  Perhaps the best fix is to re-visit those
+ * speed-up efforts, and see if a better solution can't be found.
+ *
+ * My next thing to try was to put back the correct pic_icount stuff and see if
+ * I can't set up separate HLTs for each IRQ, automatically releasing any
+ * pending (pic_pirr) interrupts when their HLT is encountered.  The pic_icount
+ * and related stuf may still be needed for stack switches.  This may be
+ * enough to gain the full speed-up without creating significant holes.
+ *
+ * The other thing I want to do is some timer changes to better guarantee
+ * interrupts get activated in their proper order.  This is a bit messy, and
+ * I've got the code somewhere. It involves threading the timer queue.  Some of
+ * it may already be done, but I think not.
+ *
+ * Larry
+ */
+
 #include <stdio.h>
 #include "config.h"
 #include "port.h"
@@ -51,7 +138,9 @@
 #else
 #include "bitops.h"
 #endif
+#include "timers.h"
 #include "pic.h"
+#include "iodev.h"
 #include "memory.h"
 #ifdef __linux__
 #include <linux/linkage.h>
@@ -60,10 +149,11 @@
 #include <sys/time.h>
 #include "cpu.h"
 #include "emu.h"
-#include "../dpmi/dpmi.h"
+#include "dpmi.h"
 #include "serial.h"
 #include "int.h"
 #include "ipx.h"
+
 #undef us
 #define us unsigned
 void timer_tick(void);
@@ -73,6 +163,17 @@ extern void timer_int_engine(void);
 static unsigned long pic1_isr;         /* second isr for pic1 irqs */
 static unsigned long pic_irq2_ivec = 0;
 
+/*
+ * pic_pirr contains a bit set for each interrupt which has attempted to
+ * re-trigger.
+ *
+ * pic_icount is an attempt to count active interrupts.  Pending (pic_pirr)
+ * interrupts are released when pic_icount gets back to zero (as designed).
+ * Haven't looked how it is now, this is one thing others have played with.
+ *
+ * pic_watch() has the primary job of releasing interrupts when a stack switch
+ * occurs.
+ */
 
 #if 0
 static unsigned long pic1_mask;        /* bits set for pic1 levels */
@@ -198,11 +299,13 @@ static unsigned char pic1_cmd;
 #ifdef NO_DEBUGPRINT_AT_ALL
 #define pic_print(code,s1,v1,s2)
 #else
-void pic_print( char code, char* s1, int v1, char* s2)
+#define pic_print(code,s1,v1,s2)	if (d.request>code){p_pic_print(s1,v1,s2);}
+
+static void p_pic_print(char *s1, int v1, char *s2)
 {
 static int oldi=0, oldc=0, header_count=0;
 char ci,cc;
-if(d.request&code){
+
   if (pic_icount > oldc) cc='+';
   else if(pic_icount < oldc) cc='-';
   else cc=' ';
@@ -213,16 +316,16 @@ if(d.request&code){
   else ci=' ';
   oldi = pic_ilevel;
   if (!header_count++)
-    ifprintf(d.request,"PIC: cnt lvl pic_isr  pic_imr  pic_irr (column headers)\n");
+    log_printf(1, "PIC: cnt lvl pic_isr  pic_imr  pic_irr (column headers)\n");
   if(header_count>15) header_count=0;
   
   if(s2)
-  ifprintf(d.request,"PIC: %c%2ld %c%2ld %08lx %08lx %08lx %s%02d%s\n",
+  log_printf(1, "PIC: %c%2ld %c%2ld %08lx %08lx %08lx %s%02d%s\n",
      cc, pic_icount, ci, pic_ilevel, pic_isr, pic_imr, pic_irr, s1, v1, s2);
   else
-  ifprintf(d.request,"PIC: %c%2ld %c%2ld %08lx %08lx %08lx %s\n",
+  log_printf(1, "PIC: %c%2ld %c%2ld %08lx %08lx %08lx %s\n",
      cc, pic_icount, ci, pic_ilevel, pic_isr, pic_imr, pic_irr, s1);
-  }
+
 }
 #endif
 
@@ -231,8 +334,8 @@ if(d.request&code){
  * Pic maintains two stacks of the current interrupt level. an internal one
  * is maintained by run_irqs, and is valid whenever the emulator code for
  * an interrupt is active.  These functions maintain an external stack,
- * winch is valid from the time the dos interrupt code is called until
- * that code has issued all necessary EOIs.  Because pic will not necessarily 
+ * which is valid from the time the dos interrupt code is called until
+ * the code has issued all necessary EOIs.  Because pic will not necessarily 
  * get control immediately after an EOI, another EOI (for another interrupt)
  * could occur.  This external stack is kept strictly synchronized with
  * the actions of the dos code to avoid any problems.  pic_push and pic_pop
@@ -300,13 +403,11 @@ unsigned char int_num;
  * They are called by the code that emulates inb and outb instructions.
  * Each function implements both ports for the pic:  pic0 is on ports
  * 0x20 and 0x21; pic1 is on ports 0xa0 and 0xa1.  These functions take
- * two arguements: a port number (0 or 1) and a value to be written.
+ * two arguments: a port number (0 or 1) and a value to be written.
  *
  * DANG_END_FUNCTION
  */
-void write_pic0(port,value)
-Bit32u port;
-Bit8u value;
+void write_pic0(ioport_t port, Bit8u value)
 {
 
 /* if port == 0 this must be either an ICW1, OCW2, or OCW3
@@ -371,9 +472,7 @@ else                              /* icw2, icw3, icw4, or mask register */
 }  
 
 
-void write_pic1(port,value)
-Bit32u port;
-Bit8u value;
+void write_pic1(ioport_t port, Bit8u value)
 {
 /* if port == 0 this must be either an ICW1, OCW2, or OCW3 */
 /* if port == 1 this must be either ICW2, ICW3, ICW4, or load IMR */
@@ -442,8 +541,7 @@ else                         /* icw2, icw3, icw4, or mask register */
  *
  * DANG_END_FUNCTION
  */
-Bit8u read_pic0(port)
-Bit32u port;
+Bit8u read_pic0(ioport_t port)
 {
   port -= 0x20;
   if(port)		return((unsigned char)get_pic0_imr());
@@ -452,8 +550,7 @@ Bit32u port;
 }
 
 
-Bit8u read_pic1(port)
-Bit32u port;
+Bit8u read_pic1(ioport_t port)
 {
   port -= 0xa0;
   if(port)		return((unsigned char)get_pic1_imr());
@@ -603,7 +700,7 @@ if ((!(REG(eflags) & VIF)) && (!in_dpmi) ) {
 #else
 if (!(REG(eflags) & VIF)) return;
 #endif
-#warning using assembly run_irqs
+#warning using assembly run_irqs,expect pic1_mask not used
   __asm__ __volatile__
   ("movl "CISH_INLINE(pic_ilevel)",%%ecx\n\t" /* get old ilevel              */
    "pushl %%ecx\n\t"                          /* save old ilevel             */
@@ -617,7 +714,7 @@ if (!(REG(eflags) & VIF)) return;
    "2:bsfl %%eax,%%ebx\n\t"                  /* 2: ebx=find_bit(int_request)*/
    "jz 3f\n\t"                                /* if(!int_request) goto 3     */
    "cmpl %%ebx,%%ecx\n\t"                     /* if(ebx > pic_ilevel)...     */
-   "jl 3f\n\t"                                /* ... goto 3                  */
+   "jle 3f\n\t"                               /* ... goto 3                  */
    "btrl %%ebx,"CISH_INLINE(pic_irr)"\n\t"    /* clear_bit(ebx,&pic_irr)     */
    "jnc 2b\n\t"                               /* if bit wasn't set, go to 2  */
    "movl "CISH_INLINE(pic_isr)",%%ecx\n\t"    /* get current pic_isr         */
@@ -653,11 +750,19 @@ if (!(REG(eflags) & VIF)) return;
  *  here and fake cs:ip to PIC_[SEG,OFF], where there is a hlt.  This makes 
  *  the irq generate a sigsegv, which calls pic_iret when it completes.
  *  pic_iret then pops the real cs:ip from the stack.
- *  This routine is RE-ENTRANT - it calls run_irqs, which
+ *  This routine is RE-ENTRANT - it calls run_irqs,
  *  which may call an interrupt routine,
  *  which may call do_irq().  Be Careful!  !!!!!!!!!!!!!!!!!! 
  *  No single interrupt is ever re-entered.
  *
+ * Callers:
+ * base/misc/ioctl.c
+ * base/keyboard/serv_8042.c
+ * base/keyboard/keyboard-server.c
+ * base/serial/ser_irq.c
+ * dosext/sound/sound.c
+ * dosext/net/net/pktnew.c
+ *
  * DANG_END_FUNCTION
  */
 int do_irq()
@@ -671,16 +776,23 @@ int do_irq()
 
     if(pic_ilevel==PIC_IRQ9)      /* unvectored irq9 just calls int 0x0a.. */
       if(!IS_REDIRECTED(intr)) {intr=0x0a;pic1_isr&= 0xffef;} /* & one EOI */
+#ifndef USE_NEW_INT
     if(IS_REDIRECTED(intr)||pic_ilevel<=PIC_IRQ1||in_dpmi)
+#endif /* not USE_NEW_INT */
     {
 #if 1 /* BUG CATCHER (if 1) */
+/* outputting more then one character here will change dynamic behave such that
+ * that we mesure the wrong thing.
+ */
 g_printf("+%d",(int)pic_ilevel);
 #endif
      if(pic_ilevel < 16) pic_push(pic_ilevel);
      if (in_dpmi) {
       run_pm_int(intr);
      } else {
+#ifndef USE_NEW_INT
       pic_cli();
+#endif /* not USE_NEW_INT */
       ssp = (unsigned char *)(LWORD(ss)<<4);
       sp = (unsigned long) LWORD(esp);
 
@@ -699,15 +811,21 @@ g_printf("+%d",(int)pic_ilevel);
      }
       pic_icount++;
       pic_wcount++;
-     pic_print(2,"Initiating irq lvl ",pic_ilevel, " in do_irq");
+
+ /* enter PIC loop - we can continue only when pic_isr has been cleared */
       while(!fatalerr && test_bit(pic_ilevel,&pic_isr))
       {
+	if (d.request>2)
+		r_printf("------ PIC: intr loop ---%06lx--%06lx----\n",
+			pic_vm86_count,pic_dpmi_count);
 	if (in_dpmi ) {
           ++pic_dpmi_count;
+	  pic_print(2, "Initiating DPMI irq lvl ", pic_ilevel, " in do_irq");
 	  run_dpmi();
 	  }
 	else {
 	  ++pic_vm86_count;
+	  pic_print(2, "Initiating VM86 irq lvl ", pic_ilevel, " in do_irq");
           run_vm86();
           }
         pic_isr &= PIC_IRQALL;    /*  levels 0 and 16-31 are Auto-EOI  */
@@ -720,7 +838,9 @@ g_printf("+%d",(int)pic_ilevel);
         pkt_check_receive(0);
 #endif
 #endif
+#ifdef USE_INT_QUEUE
         int_queue_run();        /*  delete when moved to timer stuff */
+#endif
       }
       pic_sti();
       return(0);
@@ -735,7 +855,7 @@ g_printf("+%d",(int)pic_ilevel);
  * pic_request triggers an interrupt.  There is presently no way to
  * "un-trigger" an interrupt.  The interrupt will be initiated the
  * next time pic_run is called, unless masked or superceded by a 
- * higher priority interrupt.  pic_request takes one arguement, an
+ * higher priority interrupt.  pic_request takes one argument, an
  * interrupt level, which specifies the interrupt to be triggered.
  * If that interrupt is already active, the request will be queued 
  * until all active interrupts have been completed.  The queue is
@@ -756,7 +876,7 @@ static char buf[81];
 
 #if 1		/* use this result mouse slowdown in winos2 */
   if (((pic_irr|pic_isr)&(1<<inum)) || (pic_icount>pic_icount_od))
-#else          /* this make mouse work under winos2, but sometime */
+#else          /* this makes mouse work under winos2, but sometimes */
 	       /* results in internal stack overflow  */
   if(pic_isr&(1<<inum) || pic_irr&(1<<inum))
 #endif
@@ -866,14 +986,18 @@ unsigned long pic_newirr;
  * pic_watch is a watchdog timer for pending interrupts.  If pic_iret
  * somehow fails to activate a pending interrupt request for 2 consecutive 
  * timer ticks, pic_watch will activate them anyway.  pic_watch is called
- * by timer_tick, the interval timer signal handler.
+ * ONLY by timer_tick, the interval timer signal handler, so the two functions
+ * will probably be merged.
  *
  * DANG_END_FUNCTION
  */
 inline void pic_watch(s_time)
-struct timeval* s_time;
+hitimer_u *s_time;	/* time in us, 64-bit unsigned */
 {
-int timer,t_time;
+#ifndef MONOTON_MICRO_TIMING
+int timer;
+#endif
+long long t_time;
 unsigned long pic_newirr;
 
   pic_newirr=pic_wirr&~pic_irr&~pic_isr;
@@ -883,28 +1007,42 @@ unsigned long pic_newirr;
   pic_wirr|=pic_pirr;   	  /* set a new pending list for next time */
 /*  pic_activate(); */
 
-  /*  calculate new sys_time */
+/*  calculate new sys_time
+ *  non-monoton: values are kept modulo (tick_rate*15min)
+ *  monoton:     values are kept modulo 2^32 (exactly 1 hour)
+ */
+#ifndef MONOTON_MICRO_TIMING
+#if 0	/* original code, see comments in timers.h */
   t_time = (s_time->tv_sec%900)*1193047 
            + (s_time->tv_usec*1193)/1000  /* This is usec * 1.193047 split */
            + s_time->tv_usec/21277;        /* up to fit in 32 bit integer math */
+#endif
+  t_time = nmUStoTICK(s_time->td % 900000000);
+#else /* MONOTON_MICRO_TIMING */
+  t_time = UStoTICK(s_time->td);
+#endif /* MONOTON_MICRO_TIMING */
+
   /* check for any freshly initiated timers, and sync them to s_time */
-/*  for(timer=1;timer<32;++timer) {
-      if(pic_itime[timer] < (-900*1193047) && pic_itime[timer] != NEVER)
-         pic_itime[timer] += t_time - NEVER;
-  } */
   pic_print(2,"pic_itime[1]= ",pic_itime[1]," ");
+#ifndef MONOTON_MICRO_TIMING
   /* Now check for wrap-around, and adjust all counters if needed.  Adjustment
      value is 900*1193047, or total counts in 15 min. */
   if(t_time<pic_sys_time) {
      for (timer=0;timer<33;++timer) { 
        if(pic_itime[timer]>=pic_dos_time) {
-          pic_itime[timer] -= 900*1193047;
-          pic_ltime[timer] -= 900*1193047;
+	 if (pic_itime[timer] != NEVER)
+          pic_itime[timer] -= 900*PIC_TICK_RATE;
+         if (pic_ltime[timer] != NEVER)
+          pic_ltime[timer] -= 900*PIC_TICK_RATE;
           }
        else pic_ltime[timer] = pic_itime[timer] = NEVER;
      }
   }
   pic_sys_time=t_time;
+#else  /* MONOTON_MICRO_TIMING */
+  /* wraparound now is smoothly at 1h interval */
+  pic_sys_time=t_time + (t_time == NEVER);
+#endif /* MONOTON_MICRO_TIMING */
   pic_print(2,"pic_sys_time set to ",pic_sys_time," ");
   pic_dos_time = pic_itime[32];
   if(pic_icount<=pic_icount_od) pic_activate();
@@ -946,14 +1084,22 @@ int earliest, timer, count;
    earliest = pic_sys_time;
    count = 0;
    for (timer=0; timer<32; ++timer) { 
+#ifndef MONOTON_MICRO_TIMING
       if ((pic_itime[timer] < pic_sys_time) && (pic_itime[timer] != NEVER)) {
+#else /* MONOTON_MICRO_TIMING */
+      if (((int)(pic_itime[timer] - pic_sys_time) < 0) &&
+	  (pic_itime[timer] != NEVER)) {
+#endif /* MONOTON_MICRO_TIMING */
          if( pic_itime[timer] != pic_ltime[timer]) {
-           /* if( pic_itime[timer] > pic_itime[32]) {*/
+#ifndef MONOTON_MICRO_TIMING
                if ((pic_itime[timer] < earliest) || (earliest == NEVER))
+#else /* MONOTON_MICRO_TIMING */
+               if (((int)(pic_itime[timer] - earliest) < 0) ||
+		   (earliest == NEVER))
+#endif /* MONOTON_MICRO_TIMING */
                     earliest = pic_itime[timer];
                pic_request(timer);
                ++count;
-          /*  }*/
          }
       }
    }
@@ -965,6 +1111,7 @@ int earliest, timer, count;
 /* DANG_BEGIN_FUNCTION pic_sched
  * pic_sched schedules an interrupt for activation after a designated
  * time interval.  The time measurement is in unis of 1193047/second,
+ * ( or if using MONOTON_MICRO_TIMING in units of PIT_TICK_RATE/second )
  * the same rate as the pit counters.  This is convenient for timer
  * emulation, but can also be used for pacing other functions, such as
  * serial emulation, incoming keystrokes, or video updates.  Some sample 
@@ -986,19 +1133,30 @@ int earliest, timer, count;
  * DANG_END_FUNCTION
  */
  
-void pic_sched(ilevel,interval)
-int ilevel;
-int interval;
+void pic_sched(int ilevel, int interval)
 {
   char mesg[35];
+
+ /* default for interval is 65536 (=54.9ms)
+  * There's a problem with too small time intervals - an interrupt can
+  * be continuously scheduled, without letting any time to process other
+  * code.
+  *
+  * BIG WARNING - in non-periodic timer modes pit[0].cntr goes to -1
+  *	at the end of the interval - was this the reason for the following
+  *	[1u-15sec] range check?
+  */
   if(interval > 0 && interval < 0x3fffffff) {
+     if (interval < 100) {
+       if (in_dpmi) run_dpmi();  /* we're surely out of int8 here */
+     }
      if(pic_ltime[ilevel]==NEVER) {
 	pic_itime[ilevel] = pic_itime[32] + interval;
      } else {
 	pic_itime[ilevel] = pic_itime[ilevel] + interval;
      }
   }
-  if (d.request&2) {
+  if (d.request > 2) {
     /* avoid going through sprintf for non-debugging */
     sprintf(mesg,", delay= %d.",interval);
     pic_print(2,"Scheduling lvl= ",ilevel,mesg);
@@ -1010,31 +1168,27 @@ void pic_init(void)
 {
   /* do any one-time initialization of the PIC */
   emu_iodev_t  io_device;
-  emu_hlt_t    hlt_hdlr;
 
   /* 8259 PIC (Programmable Interrupt Controller) */
   io_device.read_portb   = read_pic0;
   io_device.write_portb  = write_pic0;
   io_device.read_portw   = NULL;
   io_device.write_portw  = NULL;
-  io_device.handler_name = "8259 PIC";
+  io_device.read_portd   = NULL;
+  io_device.write_portd  = NULL;
+  io_device.handler_name = "8259 PIC0";
   io_device.start_addr   = 0x0020;
   io_device.end_addr     = 0x0021;
   io_device.irq          = EMU_NO_IRQ;
-  port_register_handler(io_device);
+  io_device.fd = -1;
+  port_register_handler(io_device, 0);
 
+  io_device.handler_name = "8259 PIC1";
   io_device.start_addr = 0x00A0;
   io_device.end_addr   = 0x00A1;
   io_device.read_portb   = read_pic1;
   io_device.write_portb  = write_pic1;
-  port_register_handler(io_device);
-
-  hlt_hdlr.name       = "PIC";
-  hlt_hdlr.start_addr = 0x0fff;
-  hlt_hdlr.end_addr   = 0x0fff;
-  hlt_hdlr.func       = (emu_hlt_func)pic_iret;
-  hlt_register_handler(hlt_hdlr);
-
+  port_register_handler(io_device, 0);
 }
 
 void pic_reset(void)
@@ -1063,8 +1217,6 @@ void pic_reset(void)
   pic_seti(PIC_IRQ0, timer_int_engine, 0);  /* do_irq0 in pic.c */
   pic_unmaski(PIC_IRQ0);
   pic_request(PIC_IRQ0);                    /* start timer */
-  pic_seti(PIC_IRQ1, do_irq1, 0);           /* do_irq1 in dosio.c   */
-  pic_unmaski(PIC_IRQ1);
   if (mice->intdrv || mice->type == MOUSE_PS2) {
     pic_seti(PIC_IMOUSE, DOSEMUMouseEvents, 0);
     pic_unmaski(PIC_IMOUSE);
