@@ -1,8 +1,14 @@
+/* 
+ * (C) Copyright 1992, ..., 1999 the "DOSEMU-Development-Team".
+ *
+ * for details see file COPYING in the DOSEMU distribution
+ */
+
 #ifdef __NetBSD__
 #define EDEADLOCK EDEADLK
 #endif
 
-
+#include <features.h>
 #include <stdio.h>
 #include <string.h>
 #include <termios.h>
@@ -28,7 +34,12 @@
 #include "netbsd_vm86.h"
 #endif
 #ifdef __linux__
+#include <signal.h>
+#if GLIBC_VERSION_CODE >= 2000
+#include <sys/vt.h>
+#else
 #include <linux/vt.h>
+#endif
 #include <linux/fd.h>
 #include <linux/hdreg.h>
 #include <sys/vm86.h>
@@ -73,7 +84,6 @@ __len));
 #include "hgc.h"
 #include "ipx.h"		/* TRB - add support for ipx */
 #include "serial.h"
-#include "keymaps.h"
 #include "int.h"
 #include "bitops.h"
 #include "pic.h"
@@ -82,9 +92,13 @@ __len));
 #include <setjmp.h>
 #endif
 
+#include "keyb_clients.h"
+
 #ifdef USE_MHPDBG
   #include "mhpdbg.h"
 #endif
+
+#include "dosemu_select.h"
   
 #ifdef SIG
 extern SillyG_t *SillyG;
@@ -173,7 +187,7 @@ io_select(fd_set fds)
       break;
 
     case -1:			/* error (not EINTR) */
-      error("ERROR: bad io_select: %s\n", strerror(errno));
+      error("bad io_select: %s\n", strerror(errno));
       break;
 
     default:			/* has at least 1 descriptor ready */
@@ -183,8 +197,15 @@ io_select(fd_set fds)
 		m_printf("MOUSE: We have data\n");
 	  pic_request(PIC_IMOUSE);
 	}
-      if (FD_ISSET(kbd_fd, &fds)) {
-	getKeys();
+      if (kbd_fd >= 0) {
+        if (FD_ISSET(kbd_fd, &fds)) {
+	   keyb_client_run();
+        }
+      }
+      {
+	extern int uhook_fdin;
+	extern void uhook_input(void);
+	if (uhook_fdin != -1) if (FD_ISSET(uhook_fdin, &fds)) uhook_input();
       }
 #ifdef USE_MHPDBG
       if (mhpdbg.fdin != -1) if (FD_ISSET(mhpdbg.fdin, &fds)) mhp_input();
