@@ -1,5 +1,5 @@
 /* 
- * (C) Copyright 1992, ..., 1998 the "DOSEMU-Development-Team".
+ * (C) Copyright 1992, ..., 1999 the "DOSEMU-Development-Team".
  *
  * for details see file COPYING in the DOSEMU distribution
  */
@@ -1491,9 +1491,10 @@ mouse_delta(int event)
 void
 mouse_event()
 {
-  if (mouse.mask & mouse_events && mouse.cs && mouse.ip) {
+  if (mouse.mask & mouse_events && (mouse.cs || mouse.ip)) {
     if(in_dpmi && !in_dpmi_dos_int 
-          && !((mouse.cs == DPMI_SEG) && ((void *)(Bit32u)mouse.ip == DPMI_mouse_callback)))
+          && !((mouse.cs == DPMI_SEG) && 
+          ((void *)(Bit32u)mouse.ip == (DPMI_OFF+HLT_OFF(DPMI_mouse_callback)))))
                   fake_pm_int();
 
     /* push iret frame on _SS:_SP. At F000:20F7 (bios.S) we get an
@@ -1524,7 +1525,7 @@ mouse_event()
 
     if (in_dpmi && 
 	(REG(cs) == DPMI_SEG) 
-	&& ((void *)REG(eip) == DPMI_mouse_callback))
+	&& ((void *)REG(eip) == (DPMI_OFF+HLT_OFF(DPMI_mouse_callback))))
 	run_pm_mouse();
     
     m_printf("MOUSE: event %d, x %d ,y %d, mx %d, my %d, b %x\n",
