@@ -1,5 +1,13 @@
+/* 
+ * All modifications in this file to the original code are
+ * (C) Copyright 1992, ..., 1999 the "DOSEMU-Development-Team".
+ *
+ * for details see file COPYING in the DOSEMU distribution
+ */
+
 /* DANG_BEGIN_MODULE
  * 
+ * REMARK
  * ser_init.c: Serial ports initialization for DOSEMU
  * Please read the README.serial file in this directory for more info!
  * 
@@ -18,13 +26,14 @@
  * This module is maintained by Mark Rejhon at these Email addresses:
  *      marky@magmacom.com
  *      ag115@freenet.carleton.ca
+ * /REMARK
  *
  * maintainer:
  * Mark Rejhon <marky@ottawa.com>
  *
  * DANG_END_MODULE
  */
-
+#include <features.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
@@ -37,7 +46,11 @@
 #include <sys/stat.h>
 #include <errno.h>
 #ifdef __linux__
+#if GLIBC_VERSION_CODE < 1000
 #include <linux/fs.h>
+#else
+#include <limits.h>
+#endif
 #endif
 #include <pwd.h>
 #ifdef __NetBSD__
@@ -53,11 +66,14 @@
 #include "serial.h"
 #include "ser_defs.h"
 #include "priv.h"
+#include "utilities.h"	/* due to getpwnam */
 
 /* See README.serial file for more information on the com[] structure 
  * The declarations for this is in ../include/serial.h
  */
+#if 0
 serial_t com[MAX_SER];
+#endif
 
 /*  Determines if the tty is already locked.  Stolen from uri-dip-3.3.7k
  *  Nice work Uri Blumenthal & Ian Lance Taylor!
@@ -101,6 +117,7 @@ static int tty_already_locked(char *nam)
  */
 static int tty_lock(char *path, int mode)
 {
+  PRIV_SAVE_AREA
   char saved_path[PATH_MAX];
   char dev_nam[20];
   struct passwd *pw;
@@ -286,7 +303,7 @@ static int ser_close(int num)
 #endif
 
 Bit8u
-com_readb(Bit32u port) {
+com_readb(ioport_t port) {
   int tmp;
   for (tmp = 0; tmp < config.num_ser; tmp++) {
     if (((u_short)(port & ~7)) == com[tmp].base_port) {
@@ -297,7 +314,7 @@ com_readb(Bit32u port) {
 }
 
 void
-com_writeb(Bit32u port, Bit8u value) {
+com_writeb(ioport_t port, Bit8u value) {
   int tmp;
   for (tmp = 0; tmp < config.num_ser; tmp++) {
     if (((u_short)(port & ~7)) == com[tmp].base_port) {
@@ -387,9 +404,12 @@ static void do_ser_init(int num)
   /* Tell the port manager that we exist and that we're alive */
   io_device.read_portw  = NULL;
   io_device.write_portw = NULL;
+  io_device.read_portd  = NULL;
+  io_device.write_portd = NULL;
   io_device.start_addr  = com[num].base_port;
   io_device.end_addr    = com[num].base_port+7;
-  io_device.irq         = EMU_NO_IRQ;
+  io_device.irq         = com[num].interrupt;
+  io_device.fd		= -1;
   switch (num) {
     case 0 :
       io_device.read_portb   = com_readb;
@@ -411,8 +431,8 @@ static void do_ser_init(int num)
       io_device.write_portb  = com_writeb;
       io_device.handler_name = "COM4";
       break;
-    }
-  port_register_handler(io_device);
+  }
+  port_register_handler(io_device, 0);
 
   /* Information about serial port added to debug file */
   s_printf("SER%d: COM%d, intlevel=%d, base=0x%x, device=%s\n", 
