@@ -1,8 +1,15 @@
+/* 
+ * (C) Copyright 1992, ..., 1999 the "DOSEMU-Development-Team".
+ *
+ * for details see file COPYING in the DOSEMU distribution
+ */
+
 /* xms.c for the DOS emulator
  *       Robert Sanders, gt8134b@prism.gatech.edu
  *
  * DANG_BEGIN_MODULE
  *
+ * REMARK
  * Currently the XMS 3.0 spec is covered in this file. XMS is fairly simple
  * as it only deals with allocating extended memory and then moving it 
  * around in specific calls. This spec also includes the allocation of UMB's,
@@ -10,6 +17,7 @@
  * returned to DOS programs via the XMS requests, or int15 fnc88 is set in
  * "/etc/dosemu.conf" via the XMS paramter.
  *
+ * /REMARK
  * DANG_END_MODULE
  *
  * NOTE: I keep the BYTE size of EMB's in a field called "size" in the EMB
@@ -33,7 +41,9 @@ static int umb_find_unused(void);
  * the 1 MEG mark.  ugly.  fix this.
  */
 
+/*
 static char RCSxms[] = "$Header: /usr/src/dosemu0.60/dosemu/RCS/xms.c,v 2.11 1995/05/06 16:25:30 root Exp root $";
+*/
 
 #define	 XMS_GET_VERSION		0x00
 #define	 XMS_ALLOCATE_HIGH_MEMORY	0x01
@@ -390,8 +400,7 @@ xms_init(void)
     return;
   }
 
-  x_printf("XMS: initializing XMS...\nRCS: %s\n%d handles\n\n",
-	   RCSxms, NUM_HANDLES);
+  x_printf("XMS: initializing XMS... %d handles\n", NUM_HANDLES);
 
   freeHMA = 1;
 
@@ -428,7 +437,6 @@ xms_control(void)
     LWORD(edx) = 1;		/* HMA exists */
     x_printf("XMS driver version. XMS Spec: %04x dosemu-XMS: %04x\n",
 	     LWORD(eax), LWORD(ebx));
-    x_printf(RCSxms);
     break;
 
   case 1:			/* Request High Memory Area */
@@ -682,13 +690,32 @@ xms_query_freemem(int api)
 static void
 xms_allocate_EMB(int api)
 {
+  unsigned long totalBytes, subtotal;
   unsigned long h;
+  unsigned long kbsize;
 
   if (api == OLDXMS)
-    x_printf("XMS alloc EMB(old) size 0x%04x\n", LWORD(edx));
+    kbsize = LWORD(edx);
   else
-    x_printf("XMS alloc EMB(new) size 0x%08lx\n", REG(edx));
+    kbsize = REG(edx);
+  x_printf("XMS alloc EMB(%s) size 0x%08lx KB\n", (api==OLDXMS)?"old":"new",kbsize);
 
+  totalBytes = 0;
+  for (h = FIRST_HANDLE; h <= NUM_HANDLES; h++) {
+    if (ValidHandle(h))
+      totalBytes += handles[h].size;
+  }
+
+  subtotal = config.xms_size - (totalBytes / 1024);
+  /* total free is max allowable XMS - the number of K already allocated */
+  if(kbsize > subtotal)
+  {
+    x_printf("XMS: out of memory (only %ldK available)\n",subtotal);
+    LWORD(eax) = 0;
+    LWORD(ebx) = 0xFFa0; /* Out of memory */
+    return;
+  }
+  
   if (!(h = FindFreeHandle(FIRST_HANDLE))) {
     x_printf("XMS: out of handles\n");
     LWORD(eax) = 0;
@@ -697,12 +724,9 @@ xms_allocate_EMB(int api)
   else {
     handles[h].num = h;
     handles[h].valid = 1;
-    if (api == OLDXMS)
-      handles[h].size = (long)(LWORD(edx) * 1024);
-    else
-      handles[h].size = (long)(REG(edx) * 1024);
+    handles[h].size = kbsize*1024;
 
-    x_printf("XMS: EMB size %d\n", (int) handles[h].size);
+    x_printf("XMS: EMB size %ld bytes\n", handles[h].size);
 
     /* I could just rely on the behavior of malloc(0) here, but
        * I'd rather not.  I'm going to interpret the XMS 3.0 spec
