@@ -1,5 +1,5 @@
 /* 
- * (C) Copyright 1992, ..., 1998 the "DOSEMU-Development-Team".
+ * (C) Copyright 1992, ..., 1999 the "DOSEMU-Development-Team".
  *
  * for details see file COPYING in the DOSEMU distribution
  */
@@ -41,14 +41,18 @@
 #include <string.h>
 #include <errno.h>
 #include <sys/time.h>
-#include <asm/byteorder.h>
-#if __GLIBC__ > 1
-#include <netinet/if_ether.h>
+#if GLIBC_VERSION_CODE >= 2000
+  #if GLIBC_VERSION_CODE == 2000
+    #include <asm/byteorder.h>
+  #endif
+  #include <netinet/if_ether.h>
 #else
-#include <linux/if_ether.h>
+  #include <asm/byteorder.h>
+  #include <linux/if_ether.h>
 #endif
 #include "libpacket.h"
 #include "dosnet.h"
+#include "dosemu_select.h"
 
 /* flag to activate use of pic by packet driver */
 #if 0
