@@ -1,3 +1,9 @@
+/* 
+ * (C) Copyright 1992, ..., 1999 the "DOSEMU-Development-Team".
+ *
+ * for details see file COPYING in the DOSEMU distribution
+ */
+
 #include "config.h"
 
 #ifdef USE_MHPDBG
@@ -6,7 +12,7 @@
 
 /* Define if we want graphics in X (of course we want :-) (root@zaphod) */
 /* WARNING: This may not work in BSD, because it was written for Linux! */
-
+#include <features.h>
 #include <stdio.h>
 #include <termios.h>
 #include <stdlib.h>
@@ -21,13 +27,10 @@
 #include <sys/times.h>
 #include <sys/time.h>
 #include <sys/types.h>
-#include <sys/times.h>
 #include <limits.h>
 #ifndef __NetBSD__
 #include <getopt.h>
 #endif
-#include <assert.h>
-
 
 #if X_GRAPHICS
 #include <sys/mman.h>           /* root@sjoerd*/
@@ -39,7 +42,11 @@
 #include "netbsd_vm86.h"
 #endif
 #ifdef __linux__
+#if GLIBC_VERSION_CODE >= 2000
+#include <sys/vt.h>
+#else
 #include <linux/vt.h>
+#endif
 #include <linux/fd.h>
 #include <linux/hdreg.h>
 #include <sys/vm86.h>
@@ -83,10 +90,6 @@
 
 #include "dma.h"
 
-#ifdef bon
-extern void     vm86_GP_fault();
-#endif
-
 /*  */
 /* vm86_GP_fault @@@  32768 MOVED_CODE_BEGIN @@@ 01/23/96, ./src/arch/linux/async/sigsegv.c --> src/emu-i386/do_vm86.c  */
 /*
@@ -105,13 +108,14 @@ void vm86_GP_fault(void)
   unsigned char *csp, *lina;
   Bit32u org_eip;
   int pref_seg;
-  int done,is_rep,is_32bit;
+  int done,is_rep,prefix66,prefix67;
 
 #if 0
   u_short *ssp;
   static int haltcount = 0;
 #endif
 
+#define LWECX	    (prefix66 ^ prefix67 ? REG(ecx) : LWORD(ecx))
 #define MAX_HALT_COUNT 3
 
 #if 0
@@ -127,12 +131,12 @@ void vm86_GP_fault(void)
 #endif
 
   if (ignore_segv) {
-    error("ERROR: sigsegv ignored!\n");
+    error("sigsegv ignored!\n");
     return;
   }
 
   if (in_sigsegv)
-    error("ERROR: in_sigsegv=%d!\n", in_sigsegv);
+    error("in_sigsegv=%d!\n", in_sigsegv);
 
   /* in_vm86 = 0; */
   in_sigsegv++;
@@ -164,7 +168,7 @@ void vm86_GP_fault(void)
 
   /* DANG_BEGIN_REMARK
    * Here we handle all prefixes prior switching to the appropriate routines
-   * The exception CS:EIP will point to the first prefix that effects the
+   * The exception CS:EIP will point to the first prefix that effects
    * the faulting instruction, hence, 0x65 0x66 is same as 0x66 0x65.
    * So we collect all prefixes and remember them.
    * - Hans Lermen
@@ -174,22 +178,21 @@ void vm86_GP_fault(void)
   #define __SEG_ADR(type, seg, reg)  type((seg << 4) + LWORD(e##reg))
   done=0;
   is_rep=0;
-  is_32bit=0;
+  prefix66=prefix67=0;
   pref_seg=-1;
 
   do {
     switch (*(csp++)) {
-       case 0x66:      /* operand prefix */  is_32bit=1; break;
+       case 0x66:      /* operand prefix */  prefix66=1; break;
+       case 0x67:      /* address prefix */  prefix67=1; break;
        case 0x2e:      /* CS */              pref_seg=REG(cs); break;
        case 0x3e:      /* DS */              pref_seg=REG(ds); break;
        case 0x26:      /* ES */              pref_seg=REG(es); break;
        case 0x36:      /* SS */              pref_seg=REG(ss); break;
        case 0x65:      /* GS */              pref_seg=REG(gs); break;
        case 0x64:      /* FS */              pref_seg=REG(fs); break;
-       case 0xf3:      /* rep */             is_rep=1; break;
-#if 0
        case 0xf2:      /* repnz */
-#endif
+       case 0xf3:      /* rep */             is_rep=1; break;
        default: done=1;
     }
   } while (!done);
@@ -200,143 +203,58 @@ void vm86_GP_fault(void)
   switch (*csp) {
 
   case 0x6c:                    /* insb */
-    /* NOTE: ES can't be overwritten */
-    if (is_rep) {
-      int delta = 1;
-      if(LWORD(eflags) &DF ) delta = -1;
-      i_printf("Doing REP F3 6C (rep insb) %04x bytes, DELTA %d\n",
-              LWORD(ecx),delta);
-      while (LWORD(ecx))  {
-        *(SEG_ADR((unsigned char *),es,di)) = inb((int) LWORD(edx));
-        LWORD(edi) += delta;
-        LWORD(ecx)--;
-      }
-    }
-    else {
-      *(SEG_ADR((unsigned char *),es,di)) = inb((int) LWORD(edx));
-      i_printf("insb(0x%04x) value %02x\n",
-              LWORD(edx),*(SEG_ADR((unsigned char *),es,di)));   
-      if(LWORD(eflags) & DF) LWORD(edi)--;
-      else LWORD(edi)++;
-    }
+    /* NOTE: ES can't be overwritten; prefixes 66,67 should use esi,edi,ecx
+     * but is anyone using extended regs in real mode? */
+    /* WARNING: no test for DI wrapping! */
+    LWORD(edi) += port_rep_inb(LWORD(edx), SEG_ADR((Bit8u *),es,di),
+	LWORD(eflags)&DF, (is_rep? LWECX:1));
+    if (is_rep) LWECX = 0;
     LWORD(eip)++;
     break;
 
   case 0x6d:			/* (rep) insw / insd */
     /* NOTE: ES can't be overwritten */
-    if (is_rep) {
-      #define ___LOCALAUX(typ,iotyp,incr,x) \
-        int delta =incr; \
-        if(LWORD(eflags) & DF) delta = -incr; \
-        i_printf("rep ins%c %04x words, DELTA %d\n", x, \
-                LWORD(ecx),delta); \
-        while(LWORD(ecx)) { \
-          *(SEG_ADR(typ,es,di))=iotyp(LWORD(edx)); \
-          LWORD(edi) += delta; \
-          LWORD(ecx)--; \
-        }
-
-      if (is_32bit) {
-                                /* rep insd */
-        ___LOCALAUX((unsigned long *),ind,4,'d')
-      }
-      else {
-                                /* rep insw */
-        ___LOCALAUX((unsigned short *),inw,2,'w')
-      }
-      #undef  ___LOCALAUX
+    /* WARNING: no test for _DI wrapping! */
+    if (prefix66) {
+      LWORD(edi) += port_rep_ind(LWORD(edx), SEG_ADR((Bit32u *),es,di),
+	LWORD(eflags)&DF, (is_rep? LWECX:1));
     }
     else {
-      #define ___LOCALAUX(typ,iotyp,incr,x) \
-        *(SEG_ADR(typ,es,di)) =iotyp((int) LWORD(edx)); \
-        i_printf("ins%c(0x%04x) value %04x \n", x, \
-                LWORD(edx),*(SEG_ADR((unsigned short *),es,di))); \
-        if(LWORD(eflags) & DF) LWORD(edi) -= incr; \
-        else LWORD(edi) +=incr;
-      
-      if (is_32bit) {
-                                /* insd */
-        ___LOCALAUX((unsigned long *),ind,4,'d')
-      }
-      else {
-                                /* insw */
-        ___LOCALAUX((unsigned short *),inw,2,'w')
-      }
-      #undef  ___LOCALAUX
+      LWORD(edi) += port_rep_inw(LWORD(edx), SEG_ADR((Bit16u *),es,di),
+	LWORD(eflags)&DF, (is_rep? LWECX:1));
     }
+    if (is_rep) LWECX = 0;
     LWORD(eip)++;
     break;
       
 
   case 0x6e:			/* (rep) outsb */
     if (pref_seg < 0) pref_seg = LWORD(ds);
-    if (is_rep) {
-      int delta = 1;
-      i_printf("untested: rep outsb\n");
-      if(LWORD(eflags) & DF) delta = -1;
-      while(LWORD(ecx)) {
-        outb(LWORD(edx), *(__SEG_ADR((unsigned char *),pref_seg,si)));
-        LWORD(esi) += delta;
-        LWORD(ecx)--;
-      }
-    }
-    else {
-      i_printf("untested: outsb port 0x%04x value %02x\n",
-              LWORD(edx),*(__SEG_ADR((unsigned char *),pref_seg,si)));
-      outb(LWORD(edx), *(__SEG_ADR((unsigned char *),pref_seg,si)));
-      if(LWORD(eflags) & DF) LWORD(esi)--;
-      else LWORD(esi)++;
-    }
+    /* WARNING: no test for _SI wrapping! */
+    LWORD(esi) += port_rep_outb(LWORD(edx), __SEG_ADR((Bit8u *),pref_seg,si),
+	LWORD(eflags)&DF, (is_rep? LWECX:1));
+    if (is_rep) LWECX = 0;
     LWORD(eip)++;
     break;
 
 
   case 0x6f:			/* (rep) outsw / outsd */
     if (pref_seg < 0) pref_seg = LWORD(ds);
-    if (is_rep) {
-      #define ___LOCALAUX(typ,iotyp,incr,x) \
-        int delta = incr; \
-        i_printf("untested: rep outs%c\n", x); \
-        if(LWORD(eflags) & DF) delta = -incr; \
-        while(LWORD(ecx)) { \
-          iotyp(LWORD(edx), *(__SEG_ADR(typ,pref_seg,si))); \
-          LWORD(esi) += delta; \
-          LWORD(ecx)--; \
-        }
-      
-      if (is_32bit) {
-                                /* rep outsd */
-        ___LOCALAUX((unsigned long *),outd,4,'d')
-      }
-      else {
-                                /* rep outsw */
-        ___LOCALAUX((unsigned short *),outw,2,'w')
-      }
-      #undef  ___LOCALAUX
+    /* WARNING: no test for _SI wrapping! */
+    if (prefix66) {
+      LWORD(esi) += port_rep_outd(LWORD(edx), __SEG_ADR((Bit32u *),pref_seg,si),
+	LWORD(eflags)&DF, (is_rep? LWECX:1));
     }
     else {
-      #define ___LOCALAUX(typ,iotyp,incr,x) \
-        i_printf("untested: outs%c port 0x%04x value %04x\n", x, \
-                LWORD(edx), (unsigned int) *(__SEG_ADR(typ,pref_seg,si))); \
-        iotyp(LWORD(edx), *(__SEG_ADR(typ,pref_seg,si))); \
-        if(LWORD(eflags) & DF ) LWORD(esi) -= incr; \
-        else LWORD(esi) +=incr;
-      
-      if (is_32bit) {
-                                /* outsd */
-        ___LOCALAUX((unsigned long *),outd,4,'d')
-      }
-      else {
-                                /* outsw */
-        ___LOCALAUX((unsigned short *),outw,2,'w')
-      }
-      #undef  ___LOCALAUX
+      LWORD(esi) += port_rep_outw(LWORD(edx), __SEG_ADR((Bit16u *),pref_seg,si),
+	LWORD(eflags)&DF, (is_rep? LWECX:1));
     } 
+    if (is_rep) LWECX = 0;
     LWORD(eip)++;
     break;
 
   case 0xe5:			/* inw xx, ind xx */
-    if (is_32bit) REG(eax) = ind((int) csp[1]);
+    if (prefix66) REG(eax) = ind((int) csp[1]);
     else LWORD(eax) = inw((int) csp[1]);
     LWORD(eip) += 2;
     break;
@@ -347,7 +265,7 @@ void vm86_GP_fault(void)
     break;
 
   case 0xed:			/* inw dx, ind dx */
-    if (is_32bit) REG(eax) = ind(LWORD(edx));
+    if (prefix66) REG(eax) = ind(LWORD(edx));
     else LWORD(eax) = inw(LWORD(edx));
     LWORD(eip) += 1;
     break;
@@ -358,7 +276,7 @@ void vm86_GP_fault(void)
     break;
 
   case 0xe7:			/* outw xx */
-    if (is_32bit) outd((int)csp[1], REG(eax));
+    if (prefix66) outd((int)csp[1], REG(eax));
     else outw((int)csp[1], LWORD(eax));
     LWORD(eip) += 2;
     break;
@@ -368,8 +286,8 @@ void vm86_GP_fault(void)
     break;
 
   case 0xef:			/* outw dx */
-    if (is_32bit) outd(REG(edx), REG(eax));
-    else outw(REG(edx), REG(eax));
+    if (prefix66) outd(REG(edx), REG(eax));
+    else outw(LWORD(edx), REG(eax));
     LWORD(eip) += 1;
     break;
   case 0xee:			/* outb dx */
@@ -379,8 +297,10 @@ void vm86_GP_fault(void)
 
   case 0xf4:			/* hlt...I use it for various things,
 		  like trapping direct jumps into the XMS function */
+#ifndef USE_NEW_INT
        /* set VIF (only if necessary) */
     if (REG(eflags) & IF_MASK) REG(eflags) |= VIF_MASK;
+#endif /* not USE_NEW_INT */
           /* return with STI if VIP was set from run_dpmi; this happens
            * if pic_count is >0 and the VIP flag in dpmi_eflags was on
            */
@@ -414,29 +334,16 @@ void vm86_GP_fault(void)
     }
     break;
 
-  case 0x0f: /* RDE hack */
-    if (csp[1] == 0x06)
-      /*
-	 CLTS - ignore
-      */
-      { LWORD(eip)+=2;
-	break;
-      }
-      
-   if (csp[1] == 0x20)
-      /*
-    mov ???,cr?? - ignore
-      */
-      { if (csp[2] == 0xC0) REG(eax)=0;  /* mov eax,cr0 */
-        LWORD(eip)+=3;
-   break;
-      }
+  case 0x0f: /* was: RDE hack, now handled in cpu.c */
+    if (!cpu_trap_0f(csp, NULL)) goto op0ferr;
+    break;
 
   case 0xf0:			/* lock */
   default:
     if (is_rep) fprintf(stderr, "Nope REP F3,CSP = 0x%04x\n", csp[0]);
     /* er, why don't we advance eip here, and
 	 why does it work??  */
+op0ferr:
     REG(eip) = org_eip;
 #ifdef USE_MHPDBG
     mhp_debug(DBG_GPF, 0, 0);
@@ -444,7 +351,7 @@ void vm86_GP_fault(void)
     error("general protection at %p: %x\n", lina,*lina);
     show_regs(__FILE__, __LINE__);
     show_ints(0, 0x33);
-    error("ERROR: SIGSEGV, protected insn...exiting!\n");
+    error("SIGSEGV, protected insn...exiting!\n");
     fatalerr = 4;
     leavedos(fatalerr);		/* shouldn't return */
   }				/* end of switch() */
@@ -574,6 +481,21 @@ run_vm86(void)
   k_printf(" IOPL: %u\n", (unsigned) ((vflags & IOPL_MASK) >> 12));
 #endif
 
+#ifdef USE_NEW_INT
+  /* sync the pic interrupt state with the flags && sync VIF & IF */
+    if (_EFLAGS & VIF) {
+      set_IF();
+    } else {
+      clear_IF();
+    }
+#endif /* USE_NEW_INT */
+
+    /* This will protect us from Mr.Norton's bugs */
+    if (_EFLAGS & (AC|ID)) {
+      g_printf("BUG: flags changed to %08x\n",_EFLAGS);
+      _EFLAGS &= ~(AC|ID);
+    }
+
     in_vm86 = 0;
     switch VM86_TYPE
 	(retval) {
@@ -610,6 +532,13 @@ run_vm86(void)
 	}
 
     handle_signals();
+
+    { /* catch user hooks here */
+    	extern int uhook_fdin;
+    	extern void uhook_poll(void);
+    	if (uhook_fdin != -1) uhook_poll();
+    }
+
 #ifdef USE_MHPDBG  
     if (mhpdbg.active) mhp_debug(DBG_POLL, 0, 0);
 #endif
@@ -624,6 +553,7 @@ run_vm86(void)
     if (iq.queued)
 	do_queued_ioctl();
     /* update the pic to reflect IEF */
+#ifndef USE_NEW_INT
     if (REG(eflags) & VIF_MASK) {
 	if (pic_iflag)
 	    pic_sti();
@@ -631,6 +561,7 @@ run_vm86(void)
 	if (!pic_iflag)
 	    pic_cli();		/* pic_iflag=0 => enabled */
     }
+#endif /* not USE_NEW_INT */
 }
 /* @@@ MOVE_END @@@ 49152 */
 
