@@ -1,8 +1,15 @@
+/* 
+ * (C) Copyright 1992, ..., 1999 the "DOSEMU-Development-Team".
+ *
+ * for details see file COPYING in the DOSEMU distribution
+ */
+
 /*
  * Extensions by Robert Sanders, 1992-93
  * 
  * DANG_BEGIN_MODULE
  * 
+ * REMARK
  * Here is where DOSEMU gets booted. From emu.c external calls are made to
  * the specific I/O systems (video/keyboard/serial/etc...) to initialize
  * them. Memory is cleared/set up and the boot sector is read from the
@@ -17,14 +24,9 @@
  * The top level function emulate() is called from dos.c by way of a dll
  * entry point.
  * 
+ * /REMARK
  * DANG_END_MODULE
  * 
- * DANG_BEGIN_CHANGELOG 
- * $Date: 1995/05/06 16:25:14 $ $Source: /home/src/dosemu0.60/RCS/emu.c,v $ $Revision: 2.38 $ $State: Exp $
- * 
- * $Log: emu.c,v $
- * 
- * DANG_END_CHANGELOG
  */
 
 
@@ -58,6 +60,7 @@
 __asm__("___START___: jmp _emulate\n");
 #endif
 
+#include <features.h>
 #include <stdio.h>
 #include <termios.h>
 #include <unistd.h>
@@ -70,6 +73,7 @@ __asm__("___START___: jmp _emulate\n");
 #include <string.h>
 #include <ctype.h>
 #include <fcntl.h>
+#include <time.h>
 #include <sys/time.h>
 #include <sys/types.h>
 #include <sys/wait.h>
@@ -85,7 +89,13 @@ __asm__("___START___: jmp _emulate\n");
 #include "netbsd_vm86.h"
 #endif
 #ifdef __linux__
+#if GLIBC_VERSION_CODE >= 2000
+#include <sys/vt.h>
+#include <sys/kd.h>
+#else
 #include <linux/vt.h>
+#include <linux/kd.h>
+#endif
 #include <linux/fd.h>
 #include <linux/hdreg.h>
 #include <sys/vm86.h>
@@ -99,10 +109,6 @@ __asm__("___START___: jmp _emulate\n");
 #include "mhpdbg.h"
 #endif
 
-#ifdef REQUIRES_VM86PLUS
-  /* Please folks, don't remove this, it's required for emusys.h (within emu.h) */
-  #define __EMUSYS_parent
-#endif
 #include "emu.h"
 
 #include "bios.h"
@@ -117,32 +123,42 @@ __asm__("___START___: jmp _emulate\n");
 #include "hgc.h"
 #include "ipx.h"		/* TRB - add support for ipx */
 #include "serial.h"
-#include "keymaps.h"
 #include "int.h"
 #include "bitops.h"
+#include "port.h"
 #include "pic.h"
 #include "dpmi.h"
+#include "priv.h"   /* for priv_init */
+#include "port.h"   /* for port_init */
+#include "pci.h"
 #ifdef __NetBSD__
 #include <setjmp.h>
 #endif
+#include "speaker.h"
+#include "utilities.h"
+#include "dos2linux.h"
+
+#include "keyb_clients.h"
+
+#ifdef USE_SBEMU
+#include "sound.h"
+#endif
 
 extern void     stdio_init(void);
 extern void     time_setting_init(void);
-extern void     tmpdir_init(void);
 extern void     low_mem_init(void);
 
 extern void     shared_memory_exit(void);
 extern void     restore_vt(u_short);
 extern void     disallocate_vt(void);
-extern void     keyboard_close(void);
 extern void     vm86_GP_fault();
 extern void     config_init(int argc, char **argv);
 extern void	timer_int_engine(void);
 extern void	disk_open(struct disk *dp);
+extern void     print_version(void);
 
 extern void io_select_init(void);
 
-static int      special_nowait = 0;
 
 void 
 boot(void)
@@ -164,7 +180,7 @@ boot(void)
 	dp = &disktab[1];
 	break;
     default:
-	error("ERROR: unexpected value for config.hdiskboot\n");
+	error("unexpected value for config.hdiskboot\n");
 	leavedos(15);
     }
 
@@ -180,11 +196,11 @@ boot(void)
         d_printf ("Booting from bootfile=%s...\n",dp->boot_name);
         bfd = open (dp->boot_name, O_RDONLY);
         if (bfd == -1) {/* Abort with error */
-            error("ERROR: Boot file %s missing\n",dp->boot_name);
+            error("Boot file %s missing\n",dp->boot_name);
             leavedos(16);
         }
         if (read(bfd, buffer, SECTOR_SIZE) != SECTOR_SIZE) {
-            error("ERROR: Failed to read exactly %d bytes from %s\n",
+            error("Failed to read exactly %d bytes from %s\n",
                   SECTOR_SIZE, dp->boot_name);
             leavedos(16);
         }
@@ -194,14 +210,14 @@ boot(void)
     if (dp->type == PARTITION) {/* we boot partition boot record, not MBR! */
 	d_printf("Booting partition boot record from part=%s....\n", dp->dev_name);
 	if (RPT_SYSCALL(read(dp->fdesc, buffer, SECTOR_SIZE)) != SECTOR_SIZE) {
-	    error("ERROR: reading partition boot sector using partition %s.\n", dp->dev_name);
+	    error("reading partition boot sector using partition %s.\n", dp->dev_name);
 	    leavedos(16);
 	}
     } else if (read_sectors(dp, buffer, 0, 0, 0, 1) != SECTOR_SIZE) {
-	error("ERROR: can't boot from %s, using harddisk\n", dp->dev_name);
+	error("can't boot from %s, using harddisk\n", dp->dev_name);
 	dp = hdisktab;
 	if (read_sectors(dp, buffer, 0, 0, 0, 1) != SECTOR_SIZE) {
-	    error("ERROR: can't boot from hard disk\n");
+	    error("can't boot from hard disk\n");
 	    leavedos(16);
 	}
     }
@@ -230,6 +246,7 @@ static inline void
 SIG_init()
 {
 #if defined(SIG) && defined(REQUIRES_VM86PLUS)
+    PRIV_SAVE_AREA
     /* Get in touch with Silly Interrupt Handling */
     if (config.sillyint) {
 	char            prio_table[] =
@@ -240,8 +257,11 @@ SIG_init()
 	for (i = 0; i < sizeof(prio_table); i++) {
 	    irq = prio_table[i];
 	    if (config.sillyint & (1 << irq)) {
-		if (vm86_plus(VM86_REQUEST_IRQ, (SIGIO << 8) | irq) < 0) {
-		} else {
+		int ret;
+		enter_priv_on();
+		ret = vm86_plus(VM86_REQUEST_IRQ, (SIGIO << 8) | irq);
+		leave_priv_setting();
+		if ( ret > 0) {
 		    g_printf("Gonna monitor the IRQ %d you requested\n", irq);
 		    sg->fd = -1;
 		    sg->irq = irq;
@@ -277,11 +297,22 @@ SIG_close()
 }
 
 static inline void 
-emumodule_init(void)
+vm86plus_init(void)
 {
 #ifdef REQUIRES_VM86PLUS
+    static inline int vm86_old(struct vm86_struct* v86)
+    {
+        int __res;
+        __asm__ __volatile__("int $0x80\n"
+        :"=a" (__res):"a" ((int)113), "b" ((int)v86));
+        return __res;
+    }
     if (!vm86_plus(VM86_PLUS_INSTALL_CHECK,0)) return;
-    fprintf(stderr, "vm86plus service not available in your kernel\n\r");
+    if (!vm86_old((void *)0xffffff01)) {
+      fprintf(stderr, "your kernel contains an older (interim) vm86plus version\n\r"
+      		      "please upgrade to an newer one\n\r");
+    }
+    else fprintf(stderr, "vm86plus service not available in your kernel\n\r");
     fflush(stdout);
     fflush(stderr);
     _exit(1);
@@ -291,11 +322,9 @@ emumodule_init(void)
 static inline void
 module_init(void)
 {
-    version_init();		/* Check the OS version */
-    emumodule_init();		/* emumodule support */
+    vm86plus_init();		/* emumodule support */
     SIG_init();			/* silly int generator support */
     memcheck_init();		/* lower 1M memory map support */
-    tmpdir_init();		/* create our temporary dir */
 }
 
 #ifdef __NetBSD__
@@ -348,44 +377,53 @@ changesegs()
  * 
  */
 #ifdef __ELF__
-void 
+int
 main(int argc, char **argv)
 #else
-void 
+int
 emulate(int argc, char **argv)
 #endif
 {
+    extern void parse_dosemu_users(void);
 #ifdef __NetBSD__
     changesegs();
 #endif
-    if (0 == geteuid()) {
-	warn("I am root\n");
-	i_am_root = 1;
-    } else {
-	if (getuid() != geteuid()) {
-	    fprintf(stderr, "Can't setuid to anyone else but root\n");
-	    exit(1);
-	}
-    }
 
-    priv_default();
+    srand(time(NULL));
+
+    /* NOW! it is safe to touch the priv code.  */
+    priv_init();  /* This must come first! */
+
+    /* This has to come next:
+     * Parse dosemu.users _before_ any argument usage to catch
+     * a not allowed user playing with overruns and kick him out
+     * Additionally, we check a non-suid root condition, if dosemu.user
+     * says so and exit if needed.
+     */
+    parse_dosemu_users();
 
     /* the transposal of (config_|stdio_)init allows the addition of -o */
     /* to specify a debug out filename, if you're wondering */
 
+    get_time_init();		/* debug can use CPUtime */
     io_select_init();
+    port_init();		/* setup port structures, before config! */
+    version_init();		/* Check the OS version */
     config_init(argc, argv);	/* parse the commands & config file(s) */
+    get_time_init();
     stdio_init();		/* initialize stdio & open debug file */
+    print_version();            /* log version information */
     module_init();
     low_mem_init();		/* initialize the lower 1Meg */
     time_setting_init();	/* get the startup time */
     signal_init();		/* initialize sig's & sig handlers */
     device_init();		/* initialize keyboard, disk, video, etc. */
     cpu_setup();		/* setup the CPU */
+    pci_setup();
     hardware_setup();		/* setup any hardware */
+    extra_port_init();		/* setup ports dependent on config */
     memory_init();		/* initialize the memory contents */
     boot();			/* read the boot sector & get moving */
-    timer_interrupt_init();	/* start sending int 8h int signals */
 
     if (not_use_sigio)
 	k_printf("Atleast 1 NON-SIGIO file handle in use.\n");
@@ -398,6 +436,7 @@ emulate(int argc, char **argv)
 #ifdef USE_MHPDBG  
     mhp_debug(DBG_INIT, 0, 0);
 #endif
+    timer_interrupt_init();	/* start sending int 8h int signals */
 
     while (!fatalerr) {
 	++pic_vm86_count;
@@ -415,7 +454,12 @@ emulate(int argc, char **argv)
 	pic_request(16);
 #endif
 #endif
+#ifdef USE_INT_QUEUE
 	int_queue_run();
+#endif
+#ifdef USE_SBEMU
+	run_sb(); /* Beat Karcher to this one .. 8-) - AM */
+#endif
     }
 
     error("error exit: (%d,0x%04x) in_sigsegv: %d ignore_segv: %d\n",
@@ -424,9 +468,12 @@ emulate(int argc, char **argv)
     sync();
     fprintf(stderr, "Not a good day to die!!!!!\n");
     leavedos(99);
+    return 0;  /* just to make gcc happy */
 }
 
+#if 0    /* disable C-A-D until someone will fix it (if really needed) */
 extern void HMA_MAP(int);
+static int      special_nowait = 0;
 
 void
 dos_ctrl_alt_del(void)
@@ -434,7 +481,7 @@ dos_ctrl_alt_del(void)
     dbug_printf("DOS ctrl-alt-del requested.  Rebooting!\n");
     HMA_MAP(1);
     time_setting_init();
-    keyboard_flags_init();
+    keyb_server_reset();
     video_config_init();
     serial_init();
     mouse_init();
@@ -453,16 +500,7 @@ dos_ctrl_alt_del(void)
     boot();
     timer_interrupt_init();	/* start sending int 8h int signals */
 }
-
-void
-dos_ctrlc(void)
-{
-    k_printf("DOS ctrl-c!\n");
-    p_dos_str("^C\n\r");	/* print ctrl-c message */
-    keybuf_clear();
-
-    do_soft_int(0x23);
-}
+#endif
 
 static void
 ign_sigs(int sig)
@@ -472,19 +510,23 @@ ign_sigs(int sig)
 
     g_printf("ERROR: signal %d received in leavedos()\n", sig);
     show_regs(__FILE__, __LINE__);
+    flush_log();
     if (sig == SIG_TIME)
 	timerints++;
     else
 	otherints++;
 
-#define LEAVEDOS_TIMEOUT (3 * FREQ)
+#define LEAVEDOS_TIMEOUT (3 * config.freq)
 #define LEAVEDOS_SIGOUT  5
     if ((timerints >= LEAVEDOS_TIMEOUT) || (otherints >= LEAVEDOS_SIGOUT)) {
-	error("ERROR: timed/signalled out in leavedos()\n");
+	error("timed/signalled out in leavedos()\n");
 	exit(1);
     }
 }
 
+int leavedos_recurse_check = 0;
+
+
 /* "graceful" shutdown */
 void
 leavedos(int sig)
@@ -493,10 +535,15 @@ leavedos(int sig)
     struct itimerval itv;
     extern int errno;
 
-    static int recurse_check = 0;
-    if (recurse_check);
-    recurse_check = 1;
-    priv_on();
+   
+    if (leavedos_recurse_check)
+      {
+       error("leavedos called recursively, forgetting the graceful exit!\n");
+       flush_log();
+       _exit(sig);
+      }
+    leavedos_recurse_check = 1;
+    warn("leavedos(%d) called - shutting down\n", sig);
 #if 1 /* BUG CATCHER */
     if (in_vm86) {
       g_printf("\nkilled while in vm86(), trying to dump DOS-registers:\n");
@@ -505,8 +552,11 @@ leavedos(int sig)
 #endif
     in_vm86 = 0;
 
-    /* remove tmpdir */
-    rmdir(tmpdir);
+    /* try to notify dosdebug */
+    {
+      extern void mhp_exit_intercept(int errcode);
+      mhp_exit_intercept(sig);
+    }
 
     itv.it_interval.tv_sec = itv.it_interval.tv_usec = 0;
     itv.it_value = itv.it_interval;
@@ -520,6 +570,21 @@ leavedos(int sig)
     SETSIG(SIGTRAP, ign_sigs);
     warn("leavedos(%d) called - shutting down\n", sig);
 
+    if (config.speaker == SPKR_EMULATED) {
+      g_printf("SPEAKER: sound off\n");
+      speaker_off();		/* turn off any sound */
+    }
+    if (config.speaker==SPKR_NATIVE) {
+       g_printf("SPEAKER: sound off\n");
+       /* Since the speaker is native hardware use port manipulation,
+	* we don't know what is actually implementing the kernel's
+	* ioctls.
+	* My port logic is acutally stolen from kd_nosound in the kernel.
+	* 		--EB 21 September 1997
+	*/
+       port_safe_outb(0x61, port_safe_inb(0x61)&0xFC); /* turn off any sound */
+    }
+
     g_printf("calling close_all_printers\n");
     close_all_printers();
 
@@ -531,6 +596,13 @@ leavedos(int sig)
     g_printf("calling mouse_close\n");
     mouse_close();
 
+#ifdef IPX
+    {
+      extern void ipx_close(void);
+      ipx_close();
+    }
+#endif
+
 #ifdef SIG
     g_printf("calling SIG_close\n");
 #endif
@@ -541,15 +613,20 @@ leavedos(int sig)
     disk_close_all();
     g_printf("calling video_close\n");
     video_close();
+   
     g_printf("calling keyboard_close\n");
-    fflush(stderr);
-    fflush(stdout);
-    keyboard_close();
+    keyb_server_close();
+    keyb_client_close();
+
 
     g_printf("calling shared memory exit\n");
     shared_memory_exit();
     g_printf("calling HMA exit\n");
     hma_exit();
+    {
+	extern void close_uhook(void);
+	close_uhook();
+    }
 #ifdef USE_MHPDBG
     g_printf("closing debugger pipes\n");
     mhp_close();
@@ -558,6 +635,17 @@ leavedos(int sig)
 	restore_vt(config.detach);
 	disallocate_vt();
     }
+    flush_log();
+
+    /* remove per process tmpdir and its contents */
+    {
+       char *command = strcatdup("rm -rf >/dev/null 2>&1 ", TMPDIR_PROCESS);
+       if (command) {
+          priv_drop(); /* drop any priviledges before running system() !! */
+          system(command);
+       }
+    }
+
     _exit(sig);
 }
 
@@ -590,39 +678,6 @@ d_ready(int fd)
 }
 #endif
 
-static inline   uid_t
-be(uid_t who)
-{
-    if (getuid() != who)
-	return setreuid(geteuid(), getuid());
-    else
-	return 0;
-}
-
-static inline   uid_t
-be_me(void)
-{
-    if (geteuid() == 0) {
-	return setreuid(geteuid(), getuid());
-	return 0;
-    } else
-	return geteuid();
-}
-
-static inline   uid_t
-be_root(void)
-{
-    if (geteuid() != 0) {
-	setreuid(geteuid(), getuid());
-	return getuid();
-    } else
-	return 0;
-}
-
-
-
-
-
 
 void
 activate(int con_num)
