@@ -1,3 +1,10 @@
+/* 
+ * All modifications in this file to the original code are
+ * (C) Copyright 1992, ..., 1999 the "DOSEMU-Development-Team".
+ *
+ * for details see file COPYING in the DOSEMU distribution
+ */
+
 /*
  * video/cirrus.c
  * $XConsortium: cir_driver.c,v 1.5 95/01/23 15:35:14 kaleb Exp $ *
@@ -151,13 +158,13 @@ static int cirrusReprogrammedMCLK = 0;
 
  */
 
-static void wrinx(const int port, const int index, const int value)
+static inline void wrinx(const ioport_t port, const int index, const u_char value)
 {
 	port_out(index, port);
 	port_out(value, port+1);
 }
 
-static int rdinx(const int port, const int index)
+static inline int rdinx(const ioport_t port, const int index)
 {
 	port_out(index, port);
 	return (port_in(port+1) & 0xff);
