@@ -1,7 +1,13 @@
+/* 
+ * (C) Copyright 1992, ..., 1999 the "DOSEMU-Development-Team".
+ *
+ * for details see file COPYING in the DOSEMU distribution
+ */
+
 extern void vga_init_et4000(void);
 
-extern u_char et4000_ext_video_port_in(int port);
-extern void et4000_ext_video_port_out(u_char value, int port); 
+extern u_char et4000_ext_video_port_in(ioport_t port);
+extern void et4000_ext_video_port_out(ioport_t port, u_char value);
 
 extern void et4000_set_bank_read(u_char bank);
 extern void et4000_set_bank_write(u_char bank);
