@@ -1,5 +1,5 @@
 /* 
- * (C) Copyright 1992, ..., 1998 the "DOSEMU-Development-Team".
+ * (C) Copyright 1992, ..., 1999 the "DOSEMU-Development-Team".
  *
  * for details see file COPYING in the DOSEMU distribution
  */
@@ -216,7 +216,8 @@ terminal_initialize()
    Video_term.update_screen = slang_update;
    
    SLang_Exit_Error_Hook = sl_exit_error;
-   if( !DOSemu_Slang_Got_Terminfo )
+   if (config.features[0]) SLtt_get_terminfo ();
+   else if( !DOSemu_Slang_Got_Terminfo )
    {
       SLtt_get_terminfo ();
       DOSemu_Slang_Got_Terminfo = 1;
@@ -333,7 +334,34 @@ v_write(int fd, unsigned char *ch, int len)
     error("(video) v_write deferred for console_video\n");
 }
 
-static char *Help[] = 
+static char *Help_o[] = 
+{
+   "NOTE: The '^@' defaults to Ctrl-^, see dosemu.conf 'terminal {escchar}' .",
+   "Function Keys:",
+   "    F1: ^@1      F2: ^@2 ...  F9: ^@9    F10: ^@0   F11: ^@-   F12: ^@=",
+   "Key Modifiers:",
+   "    ^@s : SHIFT KEY        ^@S : STICKY SHIFT KEY",
+   "    ^@a : ALT KEY          ^@A : STICKY ALT KEY",
+   "    ^@c : CTRL KEY         ^@C : STICKY CTRL KEY",
+   "  Note: To cancel the sticky key, press the sticky key again.",
+   "  Examples:",
+   "    Pressing ^@s followed by ^@3 results in SHIFT-F3.",
+   "    Pressing ^@C Up Up Up ^@C results in Ctrl-Up Ctrl-Up Ctrl-Up.",
+   "Miscellaneous:",
+   "    ^@^R : Redraw display      ^@^L : Redraw the display.",
+   "    ^@^Z : Suspend dosemu      ^@b  : Select BEST monochrome mode.",    
+   "    ^@ Up Arrow: Force the top of DOS screen to be displayed.",
+   "    ^@ Dn Arrow: Force the bottom of DOS screen to be displayed.",
+   "    ^@ Space: Reset Sticky keys and Panning to automatic panning mode.",
+   "    ^@? or ^@h:  Show this help screen.",
+   "    ^@^@:  Send the ^@ character to dos.",
+   " The default panning mode is such that the cursor will always remain visible.",
+   "",
+   "PRESS THE SPACE BAR TO CONTINUE---------",
+   NULL
+};
+
+static char *Help_n[] = 
 {
    "NOTE: The '^@' defaults to Ctrl-^, see dosemu.conf 'terminal {escchar}' .",
    "Function Keys:",
@@ -367,8 +395,11 @@ static void show_help (void)
 {
    int i;
    char *s;
+   char **Help;
+
    SLsmg_cls ();
-   
+   if (config.features[0]) Help = Help_o;
+   else Help = Help_n;
    i = 0;
    while ((s = Help[i]) != NULL) 
      {
