@@ -1,5 +1,5 @@
 /* 
- * (C) Copyright 1992, ..., 1998 the "DOSEMU-Development-Team".
+ * (C) Copyright 1992, ..., 1999 the "DOSEMU-Development-Team".
  *
  * for details see file COPYING in the DOSEMU distribution
  */
@@ -79,7 +79,7 @@
  * The update function now works for all types of modes except for
  * Hercules modes.
  * The VGA memory is now always executable.
- * -- sw (Steffen.Winterfeldt@itp.uni-leipzig.de)
+ * -- sw (Steffen Winterfeldt <wfeldt@suse.de>)
  *
  * 1997/07/07: Added linear frame buffer (LFB) support, simplified and
  * generalized video mode definitions (now possible via dosemu.conf).
@@ -104,7 +104,7 @@
 
 
 #include <features.h>
-#if __GLIBC__ == 2 && __GLIBC_MINOR__ == 0
+#if GLIBC_VERSION_CODE == 2000
 #include <sigcontext.h>
 #endif
 #include <sys/mman.h>           /* root@sjoerd*/
@@ -614,6 +614,9 @@ int vga_emu_init(vgaemu_display_type *vedt)
   emu_iodev_t io_device;
   static unsigned char *lfb_base = NULL;
 
+  vga.mode = vga.VESA_mode = 0;
+  vga.mode_info = NULL;
+
   if(config.vgaemu_memsize)
     vga.mem.size = config.vgaemu_memsize << 10;
   else
@@ -673,12 +676,6 @@ int vga_emu_init(vgaemu_display_type *vedt)
 
   vga_emu_setup_mode_table();
 
-  /* initialize other parts */
-  DAC_init();
-  Attr_init();
-  Seq_init();
-  CRTC_init();
-
   /* register VGA ports */
   io_device.read_portb = VGA_emulate_inb;
   io_device.write_portb = VGA_emulate_outb;
@@ -715,6 +712,11 @@ int vga_emu_init(vgaemu_display_type *vedt)
 
   vbe_init(vedt);
 
+  /*
+   * Set some mode; this initializes the DAC, CRTC, etc. as well.
+   */
+  vga_emu_setmode(3, 80, 25);		/* initialize some mode */
+
   v_printf(
     "VGAEmu: vga_emu_init: memory = %u kbyte at 0x%x, lfb = 0x%x\n",
     vga.mem.size >> 10, (unsigned) vga.mem.base, vga.mem.map[VGAEMU_MAP_LFB_MODE].base_page << 12
@@ -1238,11 +1240,25 @@ int vga_emu_setmode(int mode, int width, int height)
   }
 
   vga.dac.bits = 6;
-  DAC_init();		/* Re-initialize the DAC */
 
+  DAC_init();		/* Re-initialize the DAC */
+  Attr_init();
+  Seq_init();
+  CRTC_init();
+
+  /* Should be moved to Seq_init()? -- sw */
   vga.seq.chain4 = 1;	/* ??? */
   vga.seq.map_mask = 1;
 
+  if(vga.mode_info->mode >= 4 && vga.mode_info->mode <= 0x12) {
+    fprintf(
+      stderr,
+      " *** Sorry, video mode (0x%x) is not fully supported. "
+      "You may get ***\n *** better results using a recent developer release of DOSEMU.  ***\n",
+      vga.mode_info->mode
+    );
+  }
+
   return True;
 }
 
