@@ -1,3 +1,9 @@
+/* 
+ * (C) Copyright 1992, ..., 1999 the "DOSEMU-Development-Team".
+ *
+ * for details see file COPYING in the DOSEMU distribution
+ */
+
 /* dos emulator, Matthias Lautner 
  * Extensions by Robert Sanders, 1992-93
  *
@@ -6,7 +12,11 @@
 #ifndef EMU_H
 #define EMU_H
 
+#include <features.h>
 #include <sys/types.h>
+#if GLIBC_VERSION_CODE == 2000
+#include <sigcontext.h>
+#endif
 #include <signal.h> 
 #include "config.h"
 #include "types.h"
@@ -14,9 +24,6 @@
 #include "machcompat.h"
 #include "cpu.h"
 #include "vm86plus.h"
-
-/* Note: priv.h needs 'error' !!! */
-#define error(f,a...)	 	fprintf(stderr, f, ##a)
 #include "priv.h"
 
 #include "extern.h"
@@ -54,26 +61,11 @@ EXTERN int terminal_pipe;
 EXTERN int terminal_fd INIT(-1);
 EXTERN int running_kversion INIT(0);
 
-/* set to one if runing setuid as root */
-#if 1
-#define i_am_root 1  /* neadless to have it variable, it will be allways 1
-                      * If dosemu isn't suid root, it will terminate
-                      */
-#else
-EXTERN int i_am_root INIT(0);
-#endif
-
-/* set, if dosemu was started from a 'root login',
- * unset, if dosemu was started from a 'user login' (running suid root)
- */
-EXTERN int under_root_login INIT(0);
-
 EXTERN char *cstack[16384];
 
 /* this is DEBUGGING code! */
 EXTERN int sizes INIT(0);
 
-EXTERN char *tmpdir;
 EXTERN int screen, screen_mode;
 
 /* number of highest vid page - 1 */
@@ -98,9 +90,6 @@ EXTERN char *cl,		/* clear screen */
 
 /* the fd for the keyboard */ 
 EXTERN int console_fd INIT(-1);
-#ifndef NEW_KBD_CODE
-  #define kbd_fd console_fd
-#endif
 
 /* the file descriptor for /dev/mem when mmap'ing the video mem */
 EXTERN int mem_fd INIT(-1);
@@ -117,8 +106,9 @@ EXTERN int scanseq;
 EXTERN int cursor_row;
 EXTERN int cursor_col;
 
-void dos_ctrlc(void), dos_ctrl_alt_del(void);
-int ext_fs(int, char *, char *, int);
+#if 0
+void dos_ctrl_alt_del(void);	/* disabled */
+#endif
 
 EXTERN void run_vm86(void);
 EXTERN void     vm86_GP_fault();
@@ -130,128 +120,7 @@ EXTERN void     vm86_GP_fault();
 
 void getKeys(void);
 
-
-struct debug_flags {
-  unsigned char
-   disk,		/* disk msgs         "d" */
-   read,		/* disk read         "R" */
-   write,		/* disk write        "W" */
-   dos,			/* unparsed int 21h  "D" */
-   cdrom,               /* cdrom             "C" */
-   video,		/* video             "v" */
-   X,			/* X support         "X" */
-   keyb,		/* keyboard          "k" */
-   io,			/* port I/O          "i" */
-   io_trace,		/* port I/O tracing  "T" */
-   serial,		/* serial            "s" */
-   mouse,		/* mouse             "m" */
-   defint,		/* default ints      "#" */
-   printer,		/* printer           "p" */
-   general,		/* general           "g" */
-   config,		/* configuration     "c" */
-   warning,		/* warning           "w" */
-   hardware,		/* hardware          "h" */
-   IPC,			/* IPC               "I" */
-   EMS,			/* EMS               "E" */
-   xms,			/* xms               "x" */
-   dpmi,		/* dpmi              "M" */
-   network,		/* IPX network       "n" */
-   pd,			/* Packet driver     "P" */
-   request,		/* PIC               "r" */
-   sound;		/* SOUND	     "S" */
-};
-
-#if __GNUC__ >= 2
-# define FORMAT(T,A,B)  __attribute__((format(T,A,B)))
-#else
-# define FORMAT(T,A,B)
-#endif
-
-#if __GNUC__ >= 2
-# define NORETURN	__attribute__((noreturn))
-#else
-# define NORETURN
-#endif
-
-EXTERN void saytime(char *m_str);
-
-int log_printf(int, const char *,...) FORMAT(printf, 2, 3);
-
-void p_dos_str(char *,...) FORMAT(printf, 1, 2);
-
-#if 0  /* set this to 0, if you want dosemu to honor the -D flags */
- #define NO_DEBUGPRINT_AT_ALL
-#endif
-
-extern FILE *dbg_fd;
-
-#define ifprintf(flg,fmt,a...)	do{ if (flg) log_printf(flg,fmt,##a); }while(0)
-
-/* Note: moved it at top of emu.h, we need to define 'error' for priv.h 
-  #define error(f,a...)	 	fprintf(stderr, f, ##a) */
-
-#define hard_error(f, a...)	fprintf(stderr, f, ##a)
-#define dbug_printf(f,a...)	ifprintf(2,f,##a)
-#define flush_log()		{ if (dbg_fd) log_printf(-1, "\n"); }
-
-#ifndef NO_DEBUGPRINT_AT_ALL
-
-#define k_printf(f,a...) 	ifprintf(d.keyb,f,##a)
-#define h_printf(f,a...) 	ifprintf(d.hardware,f,##a)
-#define v_printf(f,a...) 	ifprintf(d.video,f,##a)
-#define X_printf(f,a...)        ifprintf(d.X,f,##a)
-#define s_printf(f,a...) 	ifprintf(d.serial,f,##a)
-#define p_printf(f,a...) 	ifprintf(d.printer,f,##a)
-#define d_printf(f,a...) 	ifprintf(d.disk,f,##a)
-#define i_printf(f,a...) 	ifprintf(d.io,f,##a)
-#define T_printf(f,a...) 	ifprintf(d.io_trace,f,##a)
-#define R_printf(f,a...) 	ifprintf(d.read,f,##a)
-#define W_printf(f,a...) 	ifprintf(d.write,f,##a)
-#define C_printf(f,a...)        ifprintf(d.cdrom,f,##a)
-#define g_printf(f,a...)	ifprintf(d.general,f,##a)
-#define x_printf(f,a...)	ifprintf(d.xms,f,##a)
-#define D_printf(f,a...)	ifprintf(d.dpmi,f,##a)
-#define m_printf(f,a...)	ifprintf(d.mouse,f,##a)
-#define I_printf(f,a...) 	ifprintf(d.IPC,f,##a)
-#define E_printf(f,a...) 	ifprintf(d.EMS,f,##a)
-#define c_printf(f,a...) 	ifprintf(d.config,f,##a)
-#define e_printf(f,a...) 	ifprintf(1,f,##a)
-#define n_printf(f,a...)        ifprintf(d.network,f,##a)  /* TRB     */
-#define pd_printf(f,a...)       ifprintf(d.pd,f,##a)	   /* pktdrvr */
-#define r_printf(f,a...)        ifprintf(d.request,f,##a)
-#define warn(f,a...)     	ifprintf(d.warning,f,##a)
-#define S_printf(f,a...)     	ifprintf(d.sound,f,##a)
-#define ds_printf(f,a...)     	ifprintf(d.dos,f,##a)
-
-#else
-
-#define k_printf(f,a...)
-#define h_printf(f,a...)
-#define v_printf(f,a...)
-#define X_printf(f,a...)
-#define s_printf(f,a...)
-#define p_printf(f,a...)
-#define d_printf(f,a...)
-#define i_printf(f,a...)
-#define R_printf(f,a...)
-#define W_printf(f,a...)
-#define C_printf(f,a...)
-#define g_printf(f,a...)
-#define x_printf(f,a...)
-#define D_printf(f,a...)
-#define m_printf(f,a...)
-#define I_printf(f,a...)
-#define E_printf(f,a...)
-#define c_printf(f,a...)
-#define e_printf(f,a...)
-#define n_printf(f,a...)
-#define pd_printf(f,a...)
-#define r_printf(f,a...)
-#define warn(f,a...)
-#define S_printf(f,a...)
-#define ds_printf(f,a...)
-
-#endif
+#include "dosemu_debug.h"
 
      void char_out(u_char, int);
 
@@ -266,29 +135,15 @@ extern FILE *dbg_fd;
 
      int set_ioperm(int, int, int);
 
-     EXTERN struct debug_flags d;
-     EXTERN u_char in_sighandler, in_ioctl;
+EXTERN struct debug_flags d INIT({0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0});
+ 
+EXTERN u_char in_sighandler, in_ioctl;
 /* one-entry queue ;-( for ioctl's */
 EXTERN struct ioctlq iq INIT({0, 0, 0, 0}); 
 EXTERN u_char in_ioctl INIT(0);
 EXTERN struct ioctlq curi INIT({0, 0, 0, 0});
 
 
-/* int 11h config single bit tests */
-#define CONF_FLOP	BIT(0)
-#define CONF_MATHCO	BIT(1)
-#define CONF_MOUSE	BIT(2)
-#define CONF_DMA	BIT(8)
-#define CONF_GAME	BIT(12)
-
-/* don't use CONF_NSER with num > 4, CONF_NLPT with num > 3, CONF_NFLOP
- * with num > 4
- */
-#define CONF_NSER(c,num)	{c&=~(BIT(9)|BIT(10)|BIT(11)); c|=(num<<9);}
-#define CONF_NLPT(c,num) 	{c&=~(BIT(14)|BIT(14)); c|=(num<<14);}
-#define CONF_NFLOP(c,num) 	{c&=~(CONF_FLOP|BIT(6)|BIT(7)); \
-				   if (num) c|=((num-1)<<6)|CONF_FLOP;}
-
 /* this macro can be safely wrapped around a system call with no side
  * effects; using a feature of GCC, it returns the same value as the
  * function call argument inside.
@@ -346,12 +201,20 @@ EXTERN struct ioctlq curi INIT({0, 0, 0, 0});
 #define False FALSE
 #endif
 
-     typedef unsigned char boolean;
+#ifndef DOSEMU_BOOL_IS_CHAR
+typedef unsigned char boolean;
+#define DOSEMU_BOOL_IS_CHAR
+#endif
 
 /* would like to have this in memory.h (but emu.h is included before memory.h !) */
 #define HARDWARE_RAM_START 0xc8000
 #define HARDWARE_RAM_STOP  0xf0000
 
+typedef struct vesamode_type_struct {
+  struct vesamode_type_struct *next;
+  unsigned width, height, color_bits;
+} vesamode_type;
+
 
      typedef struct config_info {
        int hdiskboot;
@@ -364,6 +227,8 @@ EXTERN struct ioctlq curi INIT({0, 0, 0, 0});
        boolean X;
        u_short cardtype;
        u_short chipset;
+       boolean pci;
+       boolean pci_video;
        u_short gfxmemsize;		/* for SVGA card, in K */
        /* u_short term_method; */	/* Terminal method: ANSI or NCURSES */
        u_short term_color;		/* Terminal color support on or off */
@@ -378,9 +243,23 @@ EXTERN struct ioctlq curi INIT({0, 0, 0, 0});
        char    *X_title;                /* X window title */
        char    *X_icon_name;
        char    *X_font;
+       char    *X_mgrab_key;		/* KeySym name to activate mouse grab */
+					/* "" turns it of, NULL gives the default ("Home") */
        int     X_blinkrate;
        int     X_sharecmap;
        int     X_mitshm;                /* use MIT SHM extension */
+       int     X_fixed_aspect;          /* keep initial aspect ratio while resizing windows */
+       int     X_aspect_43;             /* set aspect ratio to 4:3 */
+       int     X_lin_filt;              /* interpolate linear */
+       int     X_bilin_filt;            /* dto, bilinear */
+       int     X_winsize_x;             /* initial window width */
+       int     X_mode13fact;            /* initial size factor for mode 0x13 */
+       int     X_winsize_y;             /* initial window height */
+       unsigned X_gamma;		/* gamma correction value */
+       u_short vgaemu_memsize;		/* for VGA emulation */
+       vesamode_type *vesamode_list;	/* chained list of VESA modes */
+       int     X_lfb;			/* support VESA LFB modes */
+       int     X_pm_interface;		/* support protected mode interface */
        boolean fullrestore;
        boolean force_vt_switch;         /* in case of console_video force switch to emu VT at start */
        int     dualmon;
@@ -391,15 +270,17 @@ EXTERN struct ioctlq curi INIT({0, 0, 0, 0});
        boolean kbd_tty;
        boolean X_keycode;	/* use keycode field of event structure */
        boolean exitearly;
-       boolean mathco;
+       int     realcpu;
+       boolean mathco, smp;
        boolean ipxsup;
+       boolean vnet;
        boolean keybint;
        boolean dosbanner;
        boolean allowvideoportaccess;
+       boolean rdtsc;
        boolean timers;
        boolean mouse_flag;
        boolean mapped_bios;	/* video BIOS */
-       boolean mapped_sbios;	/* system BIOS */
        char *vbios_file;	/* loaded VBIOS file */
        boolean vbios_copy;
        int vbios_seg;           /* VGA-BIOS-segment for mapping */
@@ -421,6 +302,9 @@ EXTERN struct ioctlq curi INIT({0, 0, 0, 0});
        int pktflags;		/* global flags for packet driver */
 
        unsigned int update, freq;	/* temp timer magic */
+       unsigned int wantdelta, realdelta;
+       unsigned long cpu_spd;		/* (1/speed)<<32 */
+       unsigned long cpu_tick_spd;	/* (1.19318/speed)<<32 */
 
        unsigned int hogthreshold;
 
@@ -434,20 +318,35 @@ EXTERN struct ioctlq curi INIT({0, 0, 0, 0});
        int sillyint;            /* IRQ numbers for Silly Interrupt Generator 
        				   (bitmask, bit3..15 ==> IRQ3 .. IRQ15) */
 
-       int keyboard;
-       unsigned char *key_map;     /* pointer to the correct keyboard-map */
-       unsigned char *shift_map;
-       unsigned char *alt_map;
-       unsigned char *num_table;
+       struct keytable_entry *keytable;
+
        unsigned short detach;
        unsigned char *debugout;
        unsigned char *pre_stroke;  /* pointer to keyboard pre strokes */
+       unsigned char *pre_stroke_mem;
 
        /* Lock File business */
        char *tty_lockdir;	/* The Lock directory  */
        char *tty_lockfile;	/* Lock file pretext ie LCK.. */
        boolean tty_lockbinary;	/* Binary lock files ? */
 
+       /* list of arbitrary features
+        * (at minimum 1, will be increased when needed)
+        * The purpose of these parameters is to switch between code
+        * or code pieces that in principal do the same, but for some
+        * unknown reasons behave different between machines.
+        * If a 'features' becomes obsolete (problem solved) it will
+        * remain dummy for a while before re-used.
+        *
+        * NOTE: 'features' are not subject to permanent documentation!
+        *       They should be considered 'temporary hacks'
+        *
+        * Currently assigned:
+        *
+        *   features[0]		use new(0) or old(1) Slangcode
+        */
+       int features[1];
+
        /* Sound emulation */
        __u16 sb_base;
        __u8 sb_dma;
@@ -464,32 +363,6 @@ config_t;
 #define SPKR_NATIVE	1
 #define SPKR_EMULATED	2
 
-#define KEYB_FINNISH           0
-#define KEYB_FINNISH_LATIN1    1
-#define KEYB_US                2
-#define KEYB_UK                3
-#define KEYB_DE                4
-#define KEYB_DE_LATIN1         5
-#define KEYB_FR                6
-#define KEYB_FR_LATIN1         7
-#define KEYB_DK                8
-#define KEYB_DK_LATIN1         9
-#define KEYB_DVORAK           10
-#define KEYB_SG               11
-#define KEYB_SG_LATIN1        12
-#define KEYB_NO               13
-#define KEYB_NO_LATIN1        14
-#define KEYB_SF               15
-#define KEYB_SF_LATIN1        16
-#define KEYB_ES               17
-#define KEYB_ES_LATIN1        18
-#define KEYB_BE               19
-#define KEYB_PO               20
-#define KEYB_IT               21
-#define KEYB_SW		      22
-#define KEYB_HU		      23
-#define KEYB_HU_CWI	      24
-#define KEYB_HU_LATIN2	      25
 /*
  * Right now, dosemu only supports two serial ports.
  */
@@ -513,7 +386,12 @@ EXTERN int write_port(unsigned int, unsigned short);
 EXTERN __inline__ void parent_nextscan(void);
 EXTERN __inline__ void disk_close(void);
 EXTERN void cpu_setup(void);
-EXTERN __inline__ void run_int(int);
+EXTERN void real_run_int(int);
+#ifdef USE_NEW_INT
+  #define run_int do_int
+#else
+  #define run_int real_run_int
+#endif
 EXTERN int mfs_redirector(void);
 EXTERN void int10(void);
 EXTERN void int13(u_char);
@@ -527,7 +405,7 @@ EXTERN void floppy_tick(void);
 EXTERN void open_kmem(void);
 EXTERN void close_kmem(void);
 EXTERN void CloseNetworkLink(int);
-EXTERN int parse_config(char *);
+EXTERN int parse_config(char *, char *);
 EXTERN void disk_init(void);
 EXTERN void serial_init(void);
 EXTERN void close_all_printers(void);
@@ -539,7 +417,7 @@ EXTERN int mfs_inte6(void);
 EXTERN void pkt_helper(void);
 EXTERN short pop_word(struct vm86_regs *);
 EXTERN void ems_init(void);
-EXTERN int GetDebugFlagsHelper(char *);
+EXTERN int GetDebugFlagsHelper(char *, int);
 EXTERN int SetDebugFlagsHelper(char *);
 EXTERN void leavedos(int) NORETURN;
 EXTERN void add_to_io_select(int, unsigned char);
@@ -556,7 +434,7 @@ EXTERN void sigio(int, int, struct sigcontext *);
 
 /* signals for Linux's process control of consoles */
 #define SIG_RELEASE     SIGWINCH
-#define SIG_ACQUIRE     SIGUSR1
+#define SIG_ACQUIRE     SIGUSR2
 
  /* DANG_BEGIN_REMARK
   * We assume system call restarting... under linux 0.99pl8 and earlier,
@@ -579,14 +457,20 @@ EXTERN void sigio(int, int, struct sigcontext *);
  * description:
  *  All signals that wish to be handled properly in context with the
  * execution of vm86() mode, and signals that wish to use non-reentrant
- * functions should add themselves to the SIGNALS_THAT_QUEUE define and
- * use SETQSIG(). To that end they will also need to be set up in an
+ * functions should add themselves to the ADDSET_SIGNALS_THAT_QUEUE define
+ * and use SETQSIG(). To that end they will also need to be set up in an
  * order such as SIGIO.
  *
  * DANG_END_FUNCTION
  *
  */
-#define SIGNALS_THAT_QUEUE SIGIO|SIGALRM|SIG_RELEASE|SIG_ACQUIRE
+#define ADDSET_SIGNALS_THAT_QUEUE(x) \
+do { \
+       sigaddset(x, SIGIO); \
+       sigaddset(x, SIG_TIME); \
+       sigaddset(x, SIG_RELEASE); \
+       sigaddset(x, SIG_ACQUIRE); \
+} while(0)
 
 #ifdef __NetBSD__
 #define SignalHandler sig_t
@@ -595,7 +479,7 @@ EXTERN void sigio(int, int, struct sigcontext *);
 #define NEWSETQSIG(sig, fun)	sa.sa_handler = (__sighandler_t)fun; \
 					sa.sa_flags = SA_RESTART|SA_ONSTACK ; \
 					sigemptyset(&sa.sa_mask); \
-					sigaddset(&sa.sa_mask, SIGNALS_THAT_QUEUE); \
+					ADDSET_SIGNALS_THAT_QUEUE(&sa.sa_mask); \
 					sigaction(sig, &sa, NULL);
 
 #define SETSIG(sig, fun)	sa.sa_handler = (__sighandler_t)fun; \
@@ -612,7 +496,7 @@ EXTERN void sigio(int, int, struct sigcontext *);
 			sigaction(sig, &sa, NULL);
 #endif
 #ifdef __linux__
-#ifdef __GLIBC__
+#if GLIBC_VERSION_CODE
 #define SignalHandler __sighandler_t
 #endif
 #define NEWSETQSIG(sig, fun)	sa.sa_handler = (__sighandler_t)fun; \
@@ -622,7 +506,7 @@ EXTERN void sigio(int, int, struct sigcontext *);
 			(void (*)()) (((unsigned int)(cstack) + sizeof(cstack) - 4) & ~3); \
 					sa.sa_flags = SA_RESTART ; \
 					sigemptyset(&sa.sa_mask); \
-					sigaddset(&sa.sa_mask, SIGNALS_THAT_QUEUE); \
+					ADDSET_SIGNALS_THAT_QUEUE(&sa.sa_mask); \
 					dosemu_sigaction(sig, &sa, NULL);
 
 #define SETSIG(sig, fun)	sa.sa_handler = (SignalHandler)fun; \
@@ -664,6 +548,18 @@ EXTERN struct config_info config;
  */
 EXTERN int fatalerr INIT(0);
 
+/*
+ * DANG_BEGIN_REMARK
+ * The var 'running_DosC' is set by the DosC kernel and is used to handle
+ * some things differently, e.g. the redirector.
+ * It interfaces via INTe6,0xDC (DOS_HELPER_DOSC), but only if running_DosC
+ * is !=0. At the very startup DosC issues a INTe6,0xdcDC to set running_DosC
+ * with the contents of BX (which is the internal DosC version).
+ * DANG_END_REMARK
+ */
+EXTERN int running_DosC INIT(0);
+EXTERN int dosc_interface(void);
+
 #ifdef __NetBSD__
 #define iopl(value)			{ EXTERN int errno; if (i386_iopl(value) == -1) { g_printf("iopl failed: %s\n", strerror(errno)); leavedos(4);}}
 #endif
@@ -689,4 +585,6 @@ EXTERN void read_next_scancode_from_queue (void);
 EXTERN unsigned short detach (void);
 EXTERN void HMA_init(void);
 
+extern char path_cdrom[];
+
 #endif /* EMU_H */
