@@ -1,6 +1,6 @@
 /* 
  * All modifications in this file to the original code are
- * (C) Copyright 1992, ..., 1998 the "DOSEMU-Development-Team".
+ * (C) Copyright 1992, ..., 1999 the "DOSEMU-Development-Team".
  *
  * for details see file COPYING in the DOSEMU distribution
  */
@@ -98,8 +98,8 @@ EXTERN unsigned long pic_sp INIT(0);	       /* pointer to pic_stack */
 EXTERN unsigned long pic_rflag;        /* flag to control pic_watch */
 EXTERN unsigned long pic_vm86_count INIT(0);   /* count of times 'round the vm86 loop*/
 EXTERN unsigned long pic_dpmi_count INIT(0);   /* count of times 'round the dpmi loop*/
-EXTERN long pic_dos_time;     /* dos time of last interrupt,1193047/sec.*/
-EXTERN long pic_sys_time INIT(NEVER);     /* system time set by pic_watch */
+EXTERN long long pic_dos_time;     /* dos time of last interrupt,1193047/sec.*/
+EXTERN long long pic_sys_time INIT(NEVER);     /* system time set by pic_watch */
 /* IRQ definitions.  Each entry specifies the emulator routine to call, and
    the dos interrupt vector to use.  pic_iinfo.func is set by pic_seti(),
    as are the interrupt vectors for all but [1] through [15], which  may be
