@@ -1,5 +1,5 @@
 /* 
- * (C) Copyright 1992, ..., 1998 the "DOSEMU-Development-Team".
+ * (C) Copyright 1992, ..., 1999 the "DOSEMU-Development-Team".
  *
  * for details see file COPYING in the DOSEMU distribution
  */
@@ -7,6 +7,18 @@
 #ifndef UTILITIES_H
 #define UTILITIES_H
 
+struct cmd_db {
+	char cmdname[12];
+	void (*cmdproc)(int, char *[]);
+};
+
+int argparse(char *s, char *argvx[], int maxarg);
+typedef void cmdprintf_func(const char *fmt, ...);
+void call_cmd(const char *cmd, int maxargs, const struct cmd_db *cmdtab,
+	 cmdprintf_func *printf);
+void sigalarm_onoff(int on);
+int is_console(int fd);
+
 char *strprintable(char *s);
 char *chrprintable(char c);
 void open_proc_scan(char *name);
@@ -36,4 +48,13 @@ static __inline__ int power_of_2_sqrt(int val)
 	return 1 << (res >> 1);
 }
 
+#if (GLIBC_VERSION_CODE >= 2000) && defined(PORTABLE_BINARY)
+  #include <pwd.h>
+  #include <sys/types.h>
+  struct passwd *our_getpwuid(uid_t uid);
+  struct passwd *our_getpwnam(const char *name);
+  #define getpwnam our_getpwnam
+  #define getpwuid our_getpwuid
+#endif
+
 #endif /* UTILITIES_H */
